import unittest
import os
import tempfile
import fv3config
import fv3config.caching
from fv3config._datastore import (
    get_base_forcing_directory,
    get_orographic_forcing_directory,
    get_initial_conditions_directory,
    resolve_option,
)
from fv3config._asset_list import (
    get_orographic_forcing_asset_list,
    get_base_forcing_asset_list,
    get_initial_conditions_asset_list,
    write_asset_list,
)
from .mocks import c12_config

TEST_DIRECTORY = os.path.dirname(os.path.realpath(__file__))


DEFAULT_CONFIG = c12_config()

required_run_directory_subdirectories = ["INPUT", "RESTART"]

required_base_forcing_filenames = ["forcing_file", "grb/grb_forcing_file"]

required_orographic_forcing_filenames = ["INPUT/orographic_file"]

required_default_initial_conditions_filenames = ["INPUT/initial_conditions_file"]

required_restart_initial_conditions_filenames = (
    ["INPUT/coupler.res"]
    + ["INPUT/fv_core.res.nc"]
    + [f"INPUT/fv_core.res.tile{n}.nc" for n in range(1, 7)]
    + [f"INPUT/fv_srf_wnd.res.tile{n}.nc" for n in range(1, 7)]
    + [f"INPUT/fv_tracer.res.tile{n}.nc" for n in range(1, 7)]
    + [f"INPUT/phy_data.tile{n}.nc" for n in range(1, 7)]
    + [f"INPUT/sfc_data.tile{n}.nc" for n in range(1, 7)]
)

additional_required_filenames = [
    "data_table",
    "diag_table",
    "field_table",
    "input.nml",
    "RESTART",
    "INPUT",
]

required_config_keys = [
    "namelist",
    "diag_table",
    "data_table",
    "forcing",
    "experiment_name",
    "initial_conditions",
]

gfs_initial_conditions_dir = os.path.join(
    "fv3config-cache", "initial_conditions", "gfs_initial_conditions"
)

restart_initial_conditions_dir = os.path.join(
    "fv3config-cache", "initial_conditions", "restart_initial_conditions"
)

option_abs_path = "/nonexistent/abs/path"
option_tag = "custom_option"

empty_built_in_options_dict = {}
one_item_built_in_options_dict = {"custom_option": "/path/to/custom/option"}


class ForcingTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cache_dir = tempfile.TemporaryDirectory()
        cls.original_cache_dir = fv3config.caching.get_cache_dir()
        fv3config.caching.set_cache_dir(cls.cache_dir.name)

    @classmethod
    def tearDownClass(cls):
        fv3config.caching.set_cache_dir(cls.original_cache_dir)
        cls.cache_dir.cleanup()

    def setUp(self):
        self._run_directory_list = []

    def tearDown(self):
        for directory in self._run_directory_list:
            directory.cleanup()

    def make_run_directory(self, directory_name):
        directory = tempfile.TemporaryDirectory(directory_name)
        self._run_directory_list.append(directory)
        return directory.name

    def test_write_default_base_forcing_directory(self):
        rundir = self.make_run_directory("test_rundir")
        config = DEFAULT_CONFIG.copy()
        asset_list = get_base_forcing_asset_list(config)
        write_asset_list(asset_list, rundir)
        self.assert_subpaths_present(rundir, required_base_forcing_filenames)

    def test_write_default_orographic_forcing_directory(self):
        rundir = self.make_run_directory("test_rundir")
        config = DEFAULT_CONFIG.copy()
        asset_list = get_orographic_forcing_asset_list(config)
        write_asset_list(asset_list, rundir)
        self.assert_subpaths_present(rundir, required_orographic_forcing_filenames)

    def test_zero_resolution_orographic_forcing_directory(self):
        config = {"namelist": {"fv_core_nml": {"npx": 0, "npy": 0}}}
        with self.assertRaises(fv3config.ConfigError):
            get_orographic_forcing_directory(config)

    def test_negative_resolution_orographic_forcing_directory(self):
        config = {"namelist": {"fv_core_nml": {"npx": -1, "npy": -1}}}
        with self.assertRaises(fv3config.ConfigError):
            get_orographic_forcing_directory(config)

    def test_write_default_initial_conditions_directory(self):
        rundir = self.make_run_directory("test_rundir")
        config = DEFAULT_CONFIG.copy()
        asset_list = get_initial_conditions_asset_list(config)
        write_asset_list(asset_list, rundir)
        self.assert_subpaths_present(
            rundir, required_default_initial_conditions_filenames
        )

    def test_get_specified_initial_conditions_directory(self):
        source_rundir = self.make_run_directory("source_rundir")
        config = DEFAULT_CONFIG.copy()
        config["initial_conditions"] = source_rundir
        dirname = get_initial_conditions_directory(config)
        self.assertEqual(dirname, source_rundir)

    def test_get_bad_initial_conditions_directory(self):
        source_rundir = "/not/a/real/directory"
        config = DEFAULT_CONFIG.copy()
        config["initial_conditions"] = source_rundir
        with self.assertRaises(fv3config.ConfigError):
            get_initial_conditions_directory(config)

    def test_get_specified_forcing_directory(self):
        source_rundir = self.make_run_directory("source_rundir")
        config = DEFAULT_CONFIG.copy()
        config["forcing"] = source_rundir
        dirname = get_base_forcing_directory(config)
        self.assertEqual(dirname, source_rundir)

    def test_get_bad_forcing_directory(self):
        source_rundir = "/not/a/real/directory"
        config = DEFAULT_CONFIG.copy()
        config["forcing"] = source_rundir
        with self.assertRaises(fv3config.ConfigError):
            get_base_forcing_directory(config)

    def test_write_default_run_directory(self):
        rundir = self.make_run_directory("test_rundir")
        config = DEFAULT_CONFIG.copy()
        fv3config.write_run_directory(config, rundir)
        self.assert_subpaths_present(
            rundir,
            required_run_directory_subdirectories
            + required_default_initial_conditions_filenames
            + required_base_forcing_filenames
            + required_orographic_forcing_filenames
            + additional_required_filenames,
        )

    def test_write_run_directory_with_patch_file(self):
        sourcedir = self.make_run_directory("sourcedir")
        rundir = self.make_run_directory("test_rundir")
        config = DEFAULT_CONFIG.copy()
        open(os.path.join(sourcedir, "empty_file"), "a").close()
        config["patch_files"] = {
            "source_location": sourcedir,
            "source_name": "empty_file",
            "target_location": "",
            "target_name": "empty_file",
            "copy_method": "copy",
        }
        fv3config.write_run_directory(config, rundir)
        self.assert_subpaths_present(
            rundir,
            required_run_directory_subdirectories
            + required_default_initial_conditions_filenames
            + required_base_forcing_filenames
            + required_orographic_forcing_filenames
            + additional_required_filenames
            + ["empty_file"],
        )

    def test_restart_directory_exists_and_empty(self):
        rundir = self.make_run_directory("test_rundir")
        restart_directory = os.path.join(rundir, "RESTART")
        config = DEFAULT_CONFIG.copy()
        fv3config.write_run_directory(config, rundir)
        self.assertTrue(os.path.isdir(restart_directory))
        self.assertEqual(len(os.listdir(restart_directory)), 0)

    def test_default_config_has_required_keys(self):
        config = DEFAULT_CONFIG.copy()
        self.assertTrue(set(required_config_keys) <= set(config.keys()))

    def assert_subpaths_present(self, dirname, subpath_list):
        missing_paths = []
        for subpath in subpath_list:
            path = os.path.join(dirname, subpath)
            if not os.path.exists(path):
                missing_paths.append(path)
        self.assertTrue(len(missing_paths) == 0, missing_paths)

    def test_resolve_option_abs_path(self):
        sample_abs_path = self.make_run_directory("sample_abs_path")
        self.assertEqual(
            resolve_option(sample_abs_path, empty_built_in_options_dict),
            sample_abs_path,
        )

    def test_resolve_option_nonexistent_abs_path(self):
        with self.assertRaises(fv3config.ConfigError):
            resolve_option(option_abs_path, empty_built_in_options_dict)

    def test_resolve_option_tag_empty_built_in_options(self):
        with self.assertRaises(fv3config.ConfigError):
            resolve_option(option_tag, empty_built_in_options_dict)

    def test_resolve_option_tag_proper_built_in_options(self):
        proper_option = one_item_built_in_options_dict[option_tag]
        self.assertEqual(
            resolve_option(option_tag, one_item_built_in_options_dict), proper_option
        )


if __name__ == "__main__":
    unittest.main()



import os
import re
from datetime import timedelta
from .._exceptions import ConfigError
from .default import NAMELIST_DEFAULTS
from .._asset_list import config_to_asset_list
from ..filesystem import get_fs


def get_n_processes(config):
    n_tiles = config["namelist"]["fv_core_nml"].get(
        "ntiles", NAMELIST_DEFAULTS["ntiles"]
    )
    layout = config["namelist"]["fv_core_nml"].get(
        "layout", NAMELIST_DEFAULTS["layout"]
    )
    processors_per_tile = layout[0] * layout[1]
    return n_tiles * processors_per_tile


def get_run_duration(config):
    """Return a timedelta indicating the duration of the run.

    Args:
        config (dict): a configuration dictionary

    Returns:
        duration (timedelta): the duration of the run

    Raises:
        ValueError: if the namelist contains a non-zero value for "months"
    """
    coupler_nml = config["namelist"].get("coupler_nml", {})
    months = coupler_nml.get("months", 0)
    if months != 0:  # months have no set duration and thus cannot be timedelta
        raise ValueError(f"namelist contains non-zero value {months} for months")
    return timedelta(
        **{
            name: coupler_nml.get(name, 0)
            for name in ("seconds", "minutes", "hours", "days")
        }
    )


def get_current_date(config):
    """Return current_date from configuration dictionary. This function may read from
    the remote initial_conditions path in the given configuration dictionary.

    Args:
        config (dict): a configuration dictionary

    Returns:
        list: current_date as list of ints [year, month, day, hour, min, sec]
    """
    force_date_from_namelist = config["namelist"]["coupler_nml"].get(
        "force_date_from_namelist", False
    )
    # following code replicates the logic that the fv3gfs model uses to determine the current_date
    if force_date_from_namelist:
        current_date = config["namelist"]["coupler_nml"].get(
            "current_date", [0, 0, 0, 0, 0, 0]
        )
    else:
        coupler_res_filename = _get_coupler_res_filename(config)
        if coupler_res_filename is not None:
            current_date = _get_current_date_from_coupler_res(coupler_res_filename)
        else:
            current_date = config["namelist"]["coupler_nml"].get(
                "current_date", [0, 0, 0, 0, 0, 0]
            )
    return current_date


def _get_current_date_from_coupler_res(coupler_res_filename):
    """Return current_date specified in coupler.res file

    Args:
        coupler_res_filename (str): a coupler.res filename

    Returns:
        list: current_date as list of ints [year, month, day, hour, min, sec]
    """
    fs = get_fs(coupler_res_filename)
    with fs.open(coupler_res_filename, mode="r") as f:
        third_line = f.readlines()[2]
        current_date = [int(d) for d in re.findall(r"\d+", third_line)]
        if len(current_date) != 6:
            raise ConfigError(
                f"{coupler_res_filename} does not have a valid current model time (need six integers on third line)"
            )
    return current_date


def _get_coupler_res_filename(config):
    """Return source path for coupler.res file, if it exists in config assets."""
    asset_list = config_to_asset_list(config)
    source_path = None
    for item in asset_list:
        target_path = os.path.join(item["target_location"], item["target_name"])
        if target_path == "INPUT/coupler.res":
            if "bytes" in item:
                raise NotImplementedError(
                    "Using a bytes dict to represent a coupler.res file is not "
                    "implemented yet. Use a standard asset dict for this item."
                )
            source_path = os.path.join(item["source_location"], item["source_name"])
    return source_path


def get_timestep(config):
    """Get the model timestep from a configuration dictionary.

    Args:
        config (dict): a configuration dictionary

    Returns:
        datetime.timedelta: the model timestep
    """
    return timedelta(seconds=config["namelist"]["coupler_nml"]["dt_atmos"])



from .namelist import (
    config_to_namelist,
    config_from_namelist,
)
from .rundir import write_run_directory
from .alter import enable_restart, set_run_duration
from .derive import get_n_processes, get_run_duration, get_timestep
from .nudging import get_nudging_assets, update_config_for_nudging
from .diag_table import (
    DiagFileConfig,
    DiagFieldConfig,
    DiagTable,
    Packing,
    FileFormat,
)

from ._serialization import load, dump


def get_default_config():
    """Removed, do not use."""
    raise NotImplementedError("get_default_config has been removed")



import logging
from typing import Sequence, Union, Tuple, Mapping
import dataclasses
import datetime
from enum import Enum
import re

import dacite

from .._exceptions import ConfigError

logger = logging.getLogger("fv3config")
NUMBER_OF_TOKENS_ON_FILE_LINES = 6
NUMBER_OF_TOKENS_ON_FIELD_LINES = 8


class Packing(Enum):
    DOUBLE_PRECISION = 1
    SINGLE_PRECISION = 2


class FileFormat(Enum):
    NETCDF = 1


@dataclasses.dataclass
class DiagFieldConfig:
    """Object representing configuration for a field of a diagnostics file.
    
    Args:
        module_name: Name of Fortran module containing diagnostic.
        field_name: Name of diagnostic within Fortran code.
        output_name: Name of diagnostic to use in output NetCDF.
        time_sampling: Always set to 'all'.
        reduction_method: One of 'none', 'average', 'min', 'max'.
        regional_section: 'none' or region specification.
        packing: precision for output data.
    """

    module_name: str
    field_name: str
    output_name: str
    time_sampling: str = "all"
    reduction_method: str = "none"
    regional_section: str = "none"
    packing: Packing = Packing.SINGLE_PRECISION


@dataclasses.dataclass
class DiagFileConfig:
    """Object representing a diagnostics file configuration.
    
    Args:
        name: Name to use for NetCDF files, not including '.tile?.nc'.
        frequency: Period between records in file.
        frequency_units: One of 'years', 'months', 'days', 'hours', 'minutes', 'seconds'
        field_configs: Sequence of DiagFieldConfigs defining fields to save.
        file_format: Always FileFormat.NETCDF.
        time_axis_units: Units for time coordinate in output files. One of 'years',
            'months', 'days', 'hours', 'minutes', 'seconds'.
        time_axis_name: Name for time coordinate in output files.
    """

    name: str
    frequency: int
    frequency_units: str
    field_configs: Sequence[DiagFieldConfig]
    file_format: FileFormat = FileFormat.NETCDF
    time_axis_units: str = "hours"
    time_axis_name: str = "time"


@dataclasses.dataclass
class DiagTable:
    """Representation of diag_table, which controls Fortran diagnostics manager.

    Note:
        This implementation is based on the diag_table specification described in
    https://data1.gfdl.noaa.gov/summer-school/Lectures/July16/03_Seth1_DiagManager.pdf
        The MOM6 documentation has a useful description as well:
        https://mom6.readthedocs.io/en/latest/api/generated/pages/Diagnostics.html.

    Args:
        name: label used as attribute in output diagnostic files. Cannot contain
            spaces.
        base_time: time to be used as reference for time coordinate units.
        file_configs: sequence of DiagFileConfig's defining the diagnostics to be
            output.
    """

    name: str
    base_time: datetime.datetime
    file_configs: Sequence[DiagFileConfig]

    def __post_init__(self,):
        if " " in self.name:
            raise ConfigError(
                f"Name for diag_table cannot have spaces. Got '{self.name}'."
            )

    def __repr__(self):
        """Representation of diag_table expected by the Fortran model."""
        lines = []
        lines.append(self.name)
        lines.append(self._time_to_str(self.base_time))
        lines.append("")

        for file_ in self.file_configs:
            lines.append(self._file_repr(file_))

        lines.append("")
        for file_ in self.file_configs:
            for field in file_.field_configs:
                lines.append(self._field_repr(field, file_.name))
            lines.append("")

        return "\n".join(lines)

    def asdict(self):
        return {
            "name": self.name,
            "base_time": self.base_time,
            "file_configs": [
                dataclasses.asdict(file_, dict_factory=self._dict_with_enums_factory)
                for file_ in self.file_configs
            ],
        }

    @staticmethod
    def _dict_with_enums_factory(data):
        def convert_value(obj):
            if isinstance(obj, Enum):
                return obj.value
            return obj

        return dict((k, convert_value(v)) for k, v in data)

    def _file_repr(self, file_: DiagFileConfig) -> str:
        tokens = (
            file_.name,
            file_.frequency,
            file_.frequency_units,
            file_.file_format.value,
            file_.time_axis_units,
            file_.time_axis_name,
        )
        return ", ".join(self._token_to_str(t) for t in tokens)

    def _field_repr(self, field: DiagFieldConfig, file_name: str) -> str:
        tokens = (
            field.module_name,
            field.field_name,
            field.output_name,
            file_name,
            field.time_sampling,
            field.reduction_method,
            field.regional_section,
            field.packing.value,
        )
        return ", ".join(self._token_to_str(t) for t in tokens)

    @staticmethod
    def _time_to_str(time: datetime.datetime) -> str:
        times = [time.year, time.month, time.day, time.hour, time.minute, time.second]
        return " ".join([str(t) for t in times])

    @staticmethod
    def _str_to_time(line: str) -> datetime.datetime:
        time_sequence = [int(d) for d in re.findall(r"\d+", line)]
        return datetime.datetime(*time_sequence)

    @staticmethod
    def _str_to_token(arg: str) -> Union[str, int]:
        if arg.startswith('"') and arg.endswith('"'):
            return arg.strip('"')
        elif arg.lower() == ".true.":
            # reduction_method can use '.true.' or '"average"' for same meaning
            return "average"
        elif arg.lower() == ".false.":
            # reduction_method can use '.false.' or '"none"' for same meaning
            return "none"
        else:
            return int(arg)

    @staticmethod
    def _token_to_str(token: Union[str, int]) -> str:
        if isinstance(token, str):
            return f'"{token}"'
        else:
            return str(token)

    @staticmethod
    def _parse_line(line: str) -> Sequence[Union[str, int]]:
        token_strings = line.replace(" ", "").split("#")[0].strip(",").split(",")
        return list(map(DiagTable._str_to_token, token_strings))

    @staticmethod
    def _organize_lines(
        parsed_lines: Sequence[str],
    ) -> Tuple[Sequence[str], Mapping[str, Sequence[str]]]:
        """Separate lines into 1) a sequence of lines describe files and 2) a mapping of
        file name to sequence of lines for all fields in that file."""
        file_lines = []
        field_lines = {}
        for tokens in parsed_lines:
            if len(tokens) == NUMBER_OF_TOKENS_ON_FILE_LINES:
                file_name = tokens[0]
                file_lines.append(tokens)
                field_lines[file_name] = []
            elif len(tokens) == NUMBER_OF_TOKENS_ON_FIELD_LINES:
                file_name = tokens[3]
                if file_name not in field_lines:
                    raise ConfigError(
                        "Files must be defined before they can be used by a field in "
                        f"diag_table. {file_name} has not been defined yet."
                    )
                field_lines[file_name].append(tokens)
            else:
                logger.warning(
                    f"Ignoring a line that could not be parsed in diag_table: {tokens}"
                )
        return file_lines, field_lines

    @staticmethod
    def _construct_configs_from_lines(
        file_lines, field_lines
    ) -> Sequence[DiagFileConfig]:
        file_configs = []
        for file_tokens in file_lines:
            file_name = file_tokens[0]
            field_configs = []
            for field_tokens in field_lines[file_name]:
                field_configs.append(
                    DiagFieldConfig(
                        module_name=field_tokens[0],
                        field_name=field_tokens[1],
                        output_name=field_tokens[2],
                        time_sampling=field_tokens[4],
                        reduction_method=field_tokens[5],
                        regional_section=field_tokens[6],
                        packing=Packing(field_tokens[7]),
                    )
                )
            file_configs.append(
                DiagFileConfig(
                    name=file_name,
                    frequency=file_tokens[1],
                    frequency_units=file_tokens[2],
                    field_configs=field_configs,
                    file_format=FileFormat(file_tokens[3]),
                    time_axis_units=file_tokens[4],
                    time_axis_name=file_tokens[5],
                )
            )
        return file_configs

    @classmethod
    def from_dict(cls, diag_table: dict):
        file_configs = [
            dacite.from_dict(DiagFileConfig, f, config=dacite.Config(cast=[Enum]))
            for f in diag_table["file_configs"]
        ]
        return cls(diag_table["name"], diag_table["base_time"], file_configs)

    @classmethod
    def from_str(cls, diag_table: str):
        """Initialize DiagTable class from Fortran string representation."""
        lines = diag_table.split("\n")
        lines = [line for line in lines if line and line.strip(" ")[0] != "#"]
        name = lines[0]
        base_time = cls._str_to_time(lines[1])
        parsed_lines = list(map(cls._parse_line, lines[2:]))
        file_lines, field_lines = cls._organize_lines(parsed_lines)
        file_configs = cls._construct_configs_from_lines(file_lines, field_lines)
        return cls(name, base_time, file_configs)



import pytest
import os
import fsspec
from fsspec.implementations.memory import MemoryFileSystem
import fv3config.filesystem
import fv3config.data
from . import mocks

TEST_DIR = os.path.dirname(os.path.realpath(__file__))

MOCK_FS_FILENAMES = [
    "vcm-fv3config/data/initial_conditions/gfs_c12_example/v1.0/initial_conditions_file",
    "vcm-fv3config/data/base_forcing/v1.1/forcing_file",
    "vcm-fv3config/data/base_forcing/v1.1/grb/grb_forcing_file",
    "vcm-fv3config/data/orographic_data/v1.0/C12/orographic_file",
    "vcm-fv3config/data/gfs_nudging_data/v1.0/20160801_00.nc",
    "vcm-fv3config/data/gfs_nudging_data/v1.0/20160801_06.nc",
]
DIAG_TABLE_PATH = "gs://vcm-fv3config/config/diag_table/default/v1.0/diag_table"


def populate_mock_filesystem(fs):
    for filename in MOCK_FS_FILENAMES:
        fs.mkdir(os.path.dirname(filename))
        with fs.open(filename, mode="wb") as f:
            f.write(b"mock_data")
    with open(
        os.path.join(fv3config.data.DATA_DIR, "diag_table/diag_table_default"), "r"
    ) as f_in:
        with fs.open(DIAG_TABLE_PATH, "w") as f_out:
            f_out.write(f_in.read())


class MockGCSFileSystem(MemoryFileSystem):

    protocol = "gcs", "gs"

    def ls(self, path, recursive=False, **kwargs):
        path = self._strip_protocol(path)
        return super().ls(path, **kwargs)

    def mkdir(self, path, create_parents=True, **kwargs):
        path = self._strip_protocol(path)
        return super().mkdir(path, create_parents, **kwargs)

    def rmdir(self, path, *args, **kwargs):
        path = self._strip_protocol(path)
        return super().rmdir(path, *args, **kwargs)

    def exists(self, path):
        path = self._strip_protocol(path)
        return path in self.store or path in self.pseudo_dirs

    def open(self, path, *args, **kwargs):
        path = self._strip_protocol(path)
        return super().open(path, *args, **kwargs)


@pytest.fixture(autouse=True)
def mock_gs_fs():
    memory_fs = MockGCSFileSystem()
    populate_mock_filesystem(memory_fs)

    def mock_get_fs(path: str) -> fsspec.AbstractFileSystem:
        """Return the fsspec filesystem required to handle a given path."""
        if path.startswith("gs://"):
            return memory_fs  # mock gs storage using a memory filesystem
        else:
            return fsspec.filesystem("file")

    _get_fs, fv3config.filesystem._get_fs = fv3config.filesystem._get_fs, mock_get_fs
    yield
    fv3config.filesystem._get_fs = _get_fs


c12_config = pytest.fixture(mocks.c12_config)



import os

from .caching import get_internal_cache_dir
from .data import DATA_DIR
from ._exceptions import ConfigError
from . import filesystem


DATA_TABLE_OPTIONS = {
    "default": os.path.join(DATA_DIR, "data_table/data_table_default"),
}
DIAG_TABLE_OPTIONS = {
    "default": os.path.join(DATA_DIR, "diag_table/diag_table_default"),
    "no_output": os.path.join(DATA_DIR, "diag_table/diag_table_no_output"),
    "grid_spec": os.path.join(DATA_DIR, "diag_table/diag_table_grid_spec"),
}
DEFAULT_FIELD_TABLE_DIR = os.path.join(DATA_DIR, "field_table")
FIELD_TABLE_OPTIONS = {
    "GFDLMP": "field_table_GFDLMP",
    "ZhaoCarr": "field_table_ZhaoCarr",
}


def get_resolution(config):
    """Get the model resolution based on a configuration dictionary.

    Args:
        config (dict): a configuration dictionary

    Returns:
        resolution (str): a model resolution (e.g. 'C48' or 'C96')

    Raises:
        ConfigError: if the number of processors in x and y on a tile are unequal
    """
    npx = config["namelist"]["fv_core_nml"]["npx"]
    npy = config["namelist"]["fv_core_nml"]["npy"]
    if npx != npy:
        raise ConfigError(
            f"npx and npy in fv_core_nml must be equal, but are {npx} and {npy}"
        )
    resolution = f"C{npx-1}"
    return resolution


def get_orographic_forcing_directory(config):
    """Return the string path of the orographic forcing directory
    specified by a config dictionary.
    """
    resolution = get_resolution(config)
    if "orographic_forcing" not in config:
        raise ConfigError("config dictionary must have an 'orographic_forcing' key")
    parent_dirname = config["orographic_forcing"]
    ensure_exists(parent_dirname, "orographic_forcing")
    dirname = os.path.join(parent_dirname, resolution)
    fs = filesystem.get_fs(dirname)
    if not fs.isdir(dirname):
        valid_options = fs.listdir(parent_dirname)
        raise ConfigError(
            f"resolution {resolution} orographic forcing is not present at {dirname},"
            f" valid options are {valid_options}"
        )
    return dirname


def get_base_forcing_directory(config):
    """Return the string path of the base forcing directory
    specified by a config dictionary.
    """
    if "forcing" not in config:
        raise ConfigError("config dictionary must have a 'forcing' key")
    ensure_exists(config["forcing"], "forcing")
    return config["forcing"]


def get_initial_conditions_directory(config):
    """Return the string path of the initial conditions directory
    specified by a config dictionary.
    """
    if "initial_conditions" not in config:
        raise ConfigError("config dictionary must have an 'initial_conditions' key")
    ensure_exists(config["initial_conditions"], "initial_conditions")
    return config["initial_conditions"]


def ensure_exists(location: str, location_name: str):
    if not filesystem.get_fs(location).exists(location):
        raise ConfigError(f"{location_name} location {location} does not exist")


def check_if_data_is_downloaded():
    """Removed, do not use."""
    raise NotImplementedError("check_if_data_is_downloaded has been removed")


def ensure_data_is_downloaded():
    """Removed, do not use."""
    raise NotImplementedError("ensure_data_is_downloaded has been removed")


def refresh_downloaded_data():
    """Removed, do not use."""
    raise NotImplementedError("refresh_downloaded_data has been removed")


def resolve_option(option, built_in_options_dict):
    """Determine whether a configuration dictionary option is a built-in option or
    not and return path to file or directory representing option. An option is
    assumed to be built-in if it is not an absolute path and does not begin with gs://

    Args:
        option (str): an option
        built_in_options_dict (dict): built-in options

    Returns:
        (str): a path or url

    Raises:
        ConfigError: if option is an absolute path but does not exist or if
                     option is not in default_options_dict
    """
    if filesystem.isabs(option):
        if filesystem.get_fs(option).exists(option):
            return option
        else:
            raise ConfigError(f"The provided path {option} does not exist.")
    else:
        if option in built_in_options_dict:
            return os.path.join(get_internal_cache_dir(), built_in_options_dict[option])
        else:
            raise ConfigError(
                f"The provided option {option} is not one of the built in options: "
                f"{list(built_in_options_dict.keys())}. "
                "Paths to local files or directories must be absolute."
            )


def get_microphysics_name(config):
    """Get name of microphysics scheme from configuration dictionary

    Args:
        config (dict): a configuration dictionary

    Returns:
        str: name of microphysics scheme

    Raises:
        NotImplementedError: no microphysics name defined for specified
            imp_physics and ncld combination
    """
    imp_physics = config["namelist"]["gfs_physics_nml"].get("imp_physics")
    ncld = config["namelist"]["gfs_physics_nml"].get("ncld")
    if imp_physics == 11 and ncld == 5:
        microphysics_name = "GFDLMP"
    elif imp_physics == 99 and ncld == 1:
        microphysics_name = "ZhaoCarr"
    else:
        raise NotImplementedError(
            f"Microphysics choice imp_physics={imp_physics} and ncld={ncld} not one of the valid options"
        )
    return microphysics_name


def _return_or_infer_field_table_filename(config, field_table):
    """Return or infer the field_table filename based on the config"""
    if filesystem.get_fs(field_table).isfile(field_table):
        return field_table
    elif filesystem.get_fs(field_table).isdir(field_table):
        return _infer_field_table_filename(config, field_table)
    else:
        return field_table


def get_field_table_filename(config):
    """Get field_table filename given configuration dictionary

    Args:
        config (dict): a configuration dictionary

    Returns:
        str: field_table filename

    Raises:
        ConfigError
    """
    field_table = config.get("field_table", DEFAULT_FIELD_TABLE_DIR)
    field_table_filename = _return_or_infer_field_table_filename(config, field_table)
    if not filesystem.is_existing_absolute_path(field_table_filename):
        raise ConfigError(
            f"field_table={field_table} must either be left unset or set "
            "to an existing absolute path to a file or directory"
        )
    else:
        return field_table_filename


def _infer_field_table_filename(config, field_table_directory):
    """Infer field_table filename given configuration dictionary

    The inference is made based on settings for the microphysics.

    Args:
        config (dict): a configuration dictionary
        field_table_directory (str): a directory containing the field_table

    Returns:
        str: field_table filename

    Raises:
        NotImplementedError: if field_table for microphysics option specified
            in config has not been implemented
    """
    microphysics_name = get_microphysics_name(config)
    if microphysics_name in FIELD_TABLE_OPTIONS.keys():
        filename = FIELD_TABLE_OPTIONS[microphysics_name]
    else:
        raise NotImplementedError(
            f"Field table does not exist for {microphysics_name} microphysics"
        )
    return os.path.join(field_table_directory, filename)


def get_diag_table_filename(config):
    """Return filename for diag_table specified in config

    Args:
        config (dict): a configuration dictionary

    Returns:
        str: diag_table filename
    """
    if "diag_table" not in config:
        raise ConfigError("config dictionary must have a 'diag_table' key")
    return resolve_option(config["diag_table"], DIAG_TABLE_OPTIONS)


def get_data_table_filename(config):
    """Return filename for data_table specified in config

    Args:
        config (dict): a configuration dictionary

    Returns:
        str: data_table filename
    """
    if "data_table" not in config:
        raise ConfigError("config dictionary must have a 'data_table' key")
    return resolve_option(config["data_table"], DATA_TABLE_OPTIONS)



import fv3config
import pytest
from datetime import timedelta


@pytest.fixture
def config(c12_config):
    return c12_config


@pytest.fixture(params=["seconds", "hours", "days", "years"])
def new_duration(request):
    if request.param == "seconds":
        return timedelta(seconds=5)
    elif request.param == "hours":
        return timedelta(hours=5)
    elif request.param == "days":
        return timedelta(days=7)
    elif request.param == "years":
        return timedelta(days=365 * 10)


def test_set_run_duration_sets_duration(config, new_duration):
    new_config = fv3config.set_run_duration(config, new_duration)
    assert fv3config.get_run_duration(new_config) == new_duration


def test_set_run_duration_returns_new_dict(config, new_duration):
    new_config = fv3config.set_run_duration(config, new_duration)
    assert new_config is not config
    assert new_config["namelist"] is not config["namelist"]


def test_set_run_duration_invalid_config(config, new_duration):
    config.pop("namelist")
    with pytest.raises(fv3config.ConfigError):
        fv3config.set_run_duration(config, new_duration)



import os
import appdirs

if "FV3CONFIG_CACHE_DIR" in os.environ:
    USER_CACHE_DIR = os.environ["FV3CONFIG_CACHE_DIR"]
else:
    USER_CACHE_DIR = appdirs.user_cache_dir("fv3gfs", "vulcan")
    os.makedirs(USER_CACHE_DIR, exist_ok=True)
CACHE_PREFIX = "fv3config-cache"

CACHE_REMOTE_FILES = True


def do_remote_caching(flag: bool):
    """Set whether to cache remote files when accessed. Default is True.
    """
    if not isinstance(flag, bool):
        raise TypeError(f"flag must be a boolean, was given {flag}")
    global CACHE_REMOTE_FILES
    CACHE_REMOTE_FILES = flag


def set_cache_dir(parent_dirname):
    if not os.path.isdir(parent_dirname):
        raise ValueError(f"{parent_dirname} does not exist")
    elif not os.path.isdir(os.path.join(parent_dirname, CACHE_PREFIX)):
        os.mkdir(os.path.join(parent_dirname, CACHE_PREFIX))
    global USER_CACHE_DIR
    USER_CACHE_DIR = parent_dirname


def get_cache_dir():
    return USER_CACHE_DIR


def get_internal_cache_dir():
    return os.path.join(USER_CACHE_DIR, CACHE_PREFIX)



import io
import logging
import os

from ._datastore import (
    get_initial_conditions_directory,
    get_orographic_forcing_directory,
    get_base_forcing_directory,
)
from fv3config._datastore import (
    get_field_table_filename,
    get_diag_table_filename,
    get_data_table_filename,
)
from .config.diag_table import DiagTable
from .config._serialization import dump
from ._exceptions import ConfigError
from . import filesystem


FV3CONFIG_YML_NAME = "fv3config.yml"
logger = logging.getLogger("fv3config")


def is_dict_or_list(option):
    return isinstance(option, dict) or isinstance(option, list)


def ensure_is_list(asset):
    """If argument is a dict, return length 1 list containing dict"""
    if isinstance(asset, dict):
        return [asset]
    elif isinstance(asset, list):
        return asset
    else:
        raise ConfigError("Asset must be a dict or list of dicts")


def get_orographic_forcing_asset_list(config):
    """Return asset_list for orographic forcing"""
    source_directory = get_orographic_forcing_directory(config)
    return asset_list_from_path(
        source_directory, target_location="INPUT", copy_method="link"
    )


def get_base_forcing_asset_list(config):
    """Return asset_list for base forcing"""
    if is_dict_or_list(config["forcing"]):
        return ensure_is_list(config["forcing"])
    else:
        source_directory = get_base_forcing_directory(config)
        return asset_list_from_path(source_directory, copy_method="link")


def get_initial_conditions_asset_list(config):
    """Return asset_list for initial conditions. """
    if is_dict_or_list(config["initial_conditions"]):
        return ensure_is_list(config["initial_conditions"])
    else:
        source_directory = get_initial_conditions_directory(config)
        return asset_list_from_path(source_directory, target_location="INPUT")


def get_data_table_asset(config):
    """Return asset for data_table"""
    data_table_filename = get_data_table_filename(config)
    location, name = os.path.split(data_table_filename)
    return get_asset_dict(location, name, target_name="data_table")


def get_diag_table_asset(config):
    """Return asset for diag_table"""
    if isinstance(config["diag_table"], DiagTable):
        return get_bytes_asset_dict(
            bytes(str(config["diag_table"]), "UTF-8"), ".", "diag_table"
        )
    else:
        diag_table_filename = get_diag_table_filename(config)
        location, name = os.path.split(diag_table_filename)
        return get_asset_dict(location, name, target_name="diag_table")


def get_field_table_asset(config):
    """Return asset for field_table"""
    field_table_filename = get_field_table_filename(config)
    location, name = os.path.split(field_table_filename)
    return get_asset_dict(location, name, target_name="field_table")


def get_fv3config_yaml_asset(config):
    """An asset containing this configuration"""
    f = io.StringIO()
    dump(config, f)
    return get_bytes_asset_dict(
        bytes(f.getvalue(), "UTF-8"),
        target_location=".",
        target_name=FV3CONFIG_YML_NAME,
    )


def get_asset_dict(
    source_location,
    source_name,
    target_location="",
    target_name=None,
    copy_method="copy",
):
    """Helper function to generate asset for a particular file

    Args:
        source_location (str): path to directory containing source file
        source_name (str): filename of source file
        target_location (str, optional): sub-directory to which file will
            be written, relative to run directory root. Defaults to empty
            string (i.e. root of run directory).
        target_name (str, optional): filename to which file will be written.
            Defaults to None, in which case source_name is used.
        copy_method (str, optional): flag to determine whether file is
            copied ('copy') or hard-linked ('link'). Defaults to 'copy'.

    Returns:
        dict: an asset dictionary
    """
    if target_name is None:
        target_name = source_name
    asset = {
        "source_location": source_location,
        "source_name": source_name,
        "target_location": target_location,
        "target_name": target_name,
        "copy_method": copy_method,
    }
    return asset


def get_bytes_asset_dict(
    data: bytes, target_location: str, target_name: str,
):
    """Helper function to define the necessary fields for a binary asset to
    be saved at a given location.

    Args:
        data: the bytes to save
        target_location: sub-directory to which file will
            be written, relative to run directory root. Defaults to empty
            string (i.e. root of run directory).
        target_name: filename to which file will be written.
            Defaults to None, in which case source_name is used.

    Returns:
        dict: an asset dictionary
    """

    return {
        "bytes": data,
        "target_location": target_location,
        "target_name": target_name,
    }


def _without_dot(path):
    if path == ".":
        return ""
    else:
        return path


def asset_list_from_path(from_location, target_location="", copy_method="copy"):
    """Return asset_list from all files within a given path.

    Args:
        location (str): local path or google cloud storage url.
        target_location (str, optional): target_location used for generated assets.
            Defaults to '' which is root of run-directory.
        copy_method ('copy' or 'link', optional): whether to copy or link assets,
            defaults to 'copy'. If location is a google cloud storage url, this option
            is ignored and files are copied.

    Returns:
        list: a list of asset dictionaries
        """
    if not filesystem.is_local_path(from_location):
        copy_method = "copy"
    asset_list = []
    for dirname, basename, relative_target_location in _asset_walk(from_location):
        asset_list.append(
            get_asset_dict(
                dirname,
                basename,
                target_location=os.path.join(target_location, relative_target_location),
                copy_method=copy_method,
            )
        )
    return asset_list


def _asset_walk(location):
    fs = filesystem.get_fs(location)
    protocol_prefix = filesystem._get_protocol_prefix(location)
    path_list = fs.walk(location)
    for dirname, _, files in path_list:
        dirname = protocol_prefix + dirname
        subdir_target_location = os.path.relpath(dirname, start=location)
        for basename in files:
            yield dirname, basename, _without_dot(subdir_target_location)


def write_asset(asset, target_directory):
    """Write file represented by asset to target_directory

    Args:
        asset (dict): an asset dict
        target_directory (str): path to a directory in which all files will be written
    """

    target_path = os.path.join(
        target_directory, asset["target_location"], asset["target_name"]
    )

    os.makedirs(os.path.dirname(target_path), exist_ok=True)

    if "copy_method" in asset:
        copy_file_asset(asset, target_path)
    elif "bytes" in asset:
        logger.debug(f"Writing asset bytes to {target_path}.")
        with open(target_path, "wb") as f:
            f.write(asset["bytes"])
    else:
        raise ConfigError(
            "Cannot write asset. Asset must have either a `copy_method` or `bytes` key."
        )


def copy_file_asset(asset, target_path):
    check_asset_has_required_keys(asset)
    source_path = os.path.join(asset["source_location"], asset["source_name"])
    copy_method = asset["copy_method"]
    if copy_method == "copy":
        logger.debug(f"Copying asset from {source_path} to {target_path}.")
        filesystem.get_file(source_path, target_path)
    elif copy_method == "link":
        logger.debug(f"Linking asset from {source_path} to {target_path}.")
        link_file(source_path, target_path)
    else:
        raise ConfigError(
            f"Behavior of copy_method {copy_method} not defined for {source_path} asset"
        )


def write_assets_to_directory(config, target_directory):
    asset_list = config_to_asset_list(config)
    write_asset_list(asset_list, target_directory)


def write_asset_list(asset_list, target_directory):
    """Loop over list of assets and write them all"""
    for asset in asset_list:
        write_asset(asset, target_directory)


def check_asset_has_required_keys(asset):
    """Check asset has all of its required keys"""
    required_asset_keys = [
        "source_location",
        "source_name",
        "target_location",
        "target_name",
        "copy_method",
    ]
    for required_asset_key in required_asset_keys:
        if required_asset_key not in asset:
            raise ConfigError(f"Assets must have a {required_asset_key}")


def config_to_asset_list(config):
    """Convert a configuration dictionary to an asset list. The asset list
    will contain all files for the run directory except the namelist."""
    asset_list = get_initial_conditions_asset_list(config)
    asset_list += get_base_forcing_asset_list(config)
    asset_list += get_orographic_forcing_asset_list(config)
    asset_list.append(get_field_table_asset(config))
    asset_list.append(get_diag_table_asset(config))
    asset_list.append(get_data_table_asset(config))
    asset_list.append(get_fv3config_yaml_asset(config))
    if "patch_files" in config:
        if is_dict_or_list(config["patch_files"]):
            asset_list += ensure_is_list(config["patch_files"])
        else:
            raise ConfigError(
                "patch_files item in config dictionary must be an asset dict or "
                "list of asset dicts"
            )
    return asset_list


def link_file(source_item, target_item):
    if any(not filesystem.is_local_path(item) for item in [source_item, target_item]):
        raise NotImplementedError(
            "cannot perform linking operation involving remote urls "
            f"from {source_item} to {target_item}"
        )
    if os.path.exists(target_item):
        os.remove(target_item)
    os.symlink(source_item, target_item)



class InvalidFileError(FileNotFoundError):
    """Raised when a specified file is invalid, either non-existent or not as expected."""

    pass


class ConfigError(ValueError):
    pass


class DelayedImportError(object):
    """Mock module object which raises an ImportError or other user-specified error only
    once the module's attributes are accesssed.

    Can be used to raise ImportError only if user attempts to use optional dependencies.
    """

    def __init__(self, err):
        self.err = err

    def __getattr__(self, name):
        raise self.err



import fv3config
import io
import datetime
import pytest
import copy

diag_table_obj = fv3config.DiagTable(
    name="example_diag_table",
    base_time=datetime.datetime(2000, 1, 1),
    file_configs=[
        fv3config.DiagFileConfig(
            name="physics_diagnostics",
            frequency=1,
            frequency_units="hours",
            field_configs=[
                fv3config.DiagFieldConfig(
                    "gfs_phys", "totprcb_ave", "surface_precipitation_rate"
                ),
                fv3config.DiagFieldConfig(
                    "gfs_phys", "ULWRFtoa", "upward_longwave_radiative_flux_at_toa"
                ),
            ],
        )
    ],
)


@pytest.mark.parametrize("diag_table", [diag_table_obj, "default"])
def test_dump_load(c12_config, diag_table):
    config = copy.deepcopy(c12_config)
    config["diag_table"] = diag_table
    f = io.StringIO()
    fv3config.dump(config, f)
    f.seek(0)
    loaded = fv3config.load(f)
    assert config == loaded



import datetime
import unittest
import os
import shutil
import tempfile
import fv3config
from fv3config._asset_list import (
    is_dict_or_list,
    get_data_table_asset,
    get_diag_table_asset,
    get_field_table_asset,
    get_fv3config_yaml_asset,
    get_asset_dict,
    get_bytes_asset_dict,
    ensure_is_list,
    asset_list_from_path,
    check_asset_has_required_keys,
    write_asset,
)
import yaml
import pytest
from .mocks import c12_config


TEST_DIRECTORY = os.path.dirname(os.path.realpath(__file__))
DATA_DIRECTORY = os.path.join(os.path.dirname(TEST_DIRECTORY), "fv3config", "data")


DEFAULT_CONFIG = c12_config()


DEFAULT_DATA_TABLE_ASSET = {
    "source_location": os.path.join(DATA_DIRECTORY, "data_table"),
    "source_name": "data_table_default",
    "target_location": "",
    "target_name": "data_table",
    "copy_method": "copy",
}

DEFAULT_DIAG_TABLE_ASSET = {
    "source_location": os.path.join(DATA_DIRECTORY, "diag_table"),
    "source_name": "diag_table_default",
    "target_location": "",
    "target_name": "diag_table",
    "copy_method": "copy",
}

DEFAULT_FIELD_TABLE_ASSET = {
    "source_location": os.path.join(DATA_DIRECTORY, "field_table"),
    "source_name": "field_table_GFDLMP",
    "target_location": "",
    "target_name": "field_table",
    "copy_method": "copy",
}

SAMPLE_SOURCE_LOCATION = "source/dir"
SAMPLE_SOURCE_NAME = "filename"
SAMPLE_TARGET_LOCATION = "target/dir"
SAMPLE_TARGET_NAME = "target_filename"
SAMPLE_COPY_METHOD = "link"

SAMPLE_ASSET_NO_KWARGS = {
    "source_location": SAMPLE_SOURCE_LOCATION,
    "source_name": SAMPLE_SOURCE_NAME,
    "target_location": "",
    "target_name": SAMPLE_SOURCE_NAME,
    "copy_method": "copy",
}

SAMPLE_ASSET_WITH_KWARGS = {
    "source_location": SAMPLE_SOURCE_LOCATION,
    "source_name": SAMPLE_SOURCE_NAME,
    "target_location": SAMPLE_TARGET_LOCATION,
    "target_name": SAMPLE_TARGET_NAME,
    "copy_method": SAMPLE_COPY_METHOD,
}

TEST_SOURCE_DIR = "test_source_dir"

FILELIST_EMPTY = []

FILELIST_ONE_FILE_IN_ROOT = [
    "file_in_root_dir",
]

FILELIST_WITH_SUBDIR = [
    os.path.join("subdir", "file_in_subdir"),
    "file_in_root_dir",
]

ASSET_FILE_IN_ROOT_DIR = {
    "source_location": TEST_SOURCE_DIR,
    "source_name": "file_in_root_dir",
    "target_location": "",
    "target_name": "file_in_root_dir",
    "copy_method": "copy",
}

ASSET_FILE_IN_SUBDIR = {
    "source_location": os.path.join(TEST_SOURCE_DIR, "subdir"),
    "source_name": "file_in_subdir",
    "target_location": "subdir",
    "target_name": "file_in_subdir",
    "copy_method": "copy",
}

ASSET_LIST_EMPTY = []
ASSET_LIST_ONE_FILE_IN_ROOT = [ASSET_FILE_IN_ROOT_DIR]
ASSET_LIST_WITH_SUBDIR = [ASSET_FILE_IN_ROOT_DIR, ASSET_FILE_IN_SUBDIR]


class WorkDirectory(object):
    def __init__(self, directory_path):
        os.mkdir(directory_path)
        self.directory_path = directory_path

    def cleanup(self):
        shutil.rmtree(self.directory_path)


class AssetListTests(unittest.TestCase):

    maxDiff = 999

    def setUp(self):
        self._work_directory_list = []

    def tearDown(self):
        for directory in self._work_directory_list:
            directory.cleanup()

    def make_work_directory(self, directory_name):
        full_path = os.path.join(TEST_DIRECTORY, directory_name)
        self._work_directory_list.append(WorkDirectory(full_path))
        return full_path

    def test_is_dict_or_list(self):
        empty_dict = {}
        empty_list = []
        self.assertTrue(is_dict_or_list(empty_dict))
        self.assertTrue(is_dict_or_list(empty_list))
        self.assertFalse(is_dict_or_list("a string"))
        self.assertFalse(is_dict_or_list(1.0))
        self.assertFalse(is_dict_or_list(1))

    def test_get_data_table_asset_default(self):
        config = DEFAULT_CONFIG.copy()
        data_table_asset = get_data_table_asset(config)
        self.assertEqual(data_table_asset, DEFAULT_DATA_TABLE_ASSET)

    def test_get_diag_table_asset_default(self):
        config = DEFAULT_CONFIG.copy()
        diag_table_asset = get_diag_table_asset(config)
        self.assertEqual(diag_table_asset, DEFAULT_DIAG_TABLE_ASSET)

    def test_get_field_table_asset_default(self):
        config = DEFAULT_CONFIG.copy()
        field_table_asset = get_field_table_asset(config)
        self.assertEqual(field_table_asset, DEFAULT_FIELD_TABLE_ASSET)

    def test_get_field_table_asset_existing_filename(self):
        config = DEFAULT_CONFIG.copy()
        with tempfile.NamedTemporaryFile() as field_table:
            config["field_table"] = field_table.name
            field_table_asset = get_field_table_asset(config)
            directory, filename = os.path.split(field_table.name)
            expected_field_table_asset = get_asset_dict(
                directory, filename, target_name="field_table"
            )
            self.assertEqual(field_table_asset, expected_field_table_asset)

    def test_get_field_table_asset_non_existent_relative_path(self):
        config = DEFAULT_CONFIG.copy()
        config["field_table"] = "foo"
        with self.assertRaises(fv3config.ConfigError):
            get_field_table_asset(config)

    def test_get_field_table_asset_non_existent_absolute_path(self):
        config = DEFAULT_CONFIG.copy()
        config["field_table"] = "/foo"
        with self.assertRaises(fv3config.ConfigError):
            get_field_table_asset(config)

    def test_get_field_table_asset_existing_directory(self):
        config = DEFAULT_CONFIG.copy()
        with tempfile.TemporaryDirectory() as directory:
            default_field_table_name = DEFAULT_FIELD_TABLE_ASSET["source_name"]
            open(os.path.join(directory, default_field_table_name), "a").close()
            config["field_table"] = directory
            field_table_asset = get_field_table_asset(config)
            expected_field_table_asset = get_asset_dict(
                directory, default_field_table_name, target_name="field_table"
            )
            self.assertEqual(field_table_asset, expected_field_table_asset)

    def test_get_field_table_asset_existing_directory_absent_file(self):
        config = DEFAULT_CONFIG.copy()
        with tempfile.TemporaryDirectory() as directory:
            config["field_table"] = directory
            with self.assertRaises(fv3config.ConfigError):
                get_field_table_asset(config)

    def test_ensure_is_list(self):
        sample_dict = {}
        sample_list = []
        sample_float = 1.0
        self.assertIsInstance(ensure_is_list(sample_dict), list)
        self.assertIsInstance(ensure_is_list(sample_list), list)
        with self.assertRaises(fv3config.ConfigError):
            ensure_is_list(sample_float)

    def test_get_asset_dict_default_options(self):
        test_asset = get_asset_dict(SAMPLE_SOURCE_LOCATION, SAMPLE_SOURCE_NAME)
        self.assertEqual(test_asset, SAMPLE_ASSET_NO_KWARGS)

    def test_get_asset_dict_custom_options(self):
        test_asset = get_asset_dict(
            SAMPLE_SOURCE_LOCATION,
            SAMPLE_SOURCE_NAME,
            target_location=SAMPLE_TARGET_LOCATION,
            target_name=SAMPLE_TARGET_NAME,
            copy_method=SAMPLE_COPY_METHOD,
        )
        self.assertEqual(test_asset, SAMPLE_ASSET_WITH_KWARGS)

    def test_asset_list_from_local_dir_empty(self):
        workdir = self.make_work_directory("workdir_empty")
        self.make_empty_files(workdir, FILELIST_EMPTY)
        asset_list = asset_list_from_path(workdir)
        self.assertEqual(asset_list, ASSET_LIST_EMPTY)

    def test_asset_list_from_local_dir_one_file_in_root(self):
        workdir = self.make_work_directory("workdir_with_files")
        self.make_empty_files(
            os.path.join(workdir, TEST_SOURCE_DIR), FILELIST_ONE_FILE_IN_ROOT
        )
        asset_list = asset_list_from_path(os.path.join(workdir, TEST_SOURCE_DIR))
        # necessary to have full path for source location in test assets
        for asset in ASSET_LIST_ONE_FILE_IN_ROOT:
            asset["source_location"] = os.path.join(workdir, asset["source_location"])
        self.assertEqual(asset_list, ASSET_LIST_ONE_FILE_IN_ROOT)

    def test_asset_list_from_local_dir_with_subdir(self):
        workdir = self.make_work_directory("workdir_with_files")
        self.make_empty_files(
            os.path.join(workdir, TEST_SOURCE_DIR), FILELIST_WITH_SUBDIR
        )
        asset_list = asset_list_from_path(os.path.join(workdir, TEST_SOURCE_DIR))
        # necessary to have full path for source location in test assets
        for asset in ASSET_LIST_WITH_SUBDIR:
            asset["source_location"] = os.path.join(workdir, asset["source_location"])
        self.assertEqual(asset_list, ASSET_LIST_WITH_SUBDIR)

    @staticmethod
    def make_empty_files(directory, filelist):
        for path in filelist:
            full_path = os.path.join(directory, path)
            head, tail = os.path.split(full_path)
            if not os.path.exists(head):
                os.makedirs(head)
            open(full_path, "a").close()

    def test_check_asset_has_required_keys_bad_asset(self):
        bad_asset_dict = {"irrelevant_key": ""}
        with self.assertRaises(fv3config.ConfigError):
            check_asset_has_required_keys(bad_asset_dict)

    def test_check_asset_has_required_keys_proper_asset(self):
        proper_asset_dict = {
            "source_location": "",
            "source_name": "",
            "target_location": "",
            "target_name": "",
            "copy_method": "",
        }
        try:
            check_asset_has_required_keys(proper_asset_dict)
        except fv3config.ConfigError:
            self.fail("check_asset_valid raised ConfigError unexpectedly")

    def test_write_asset(self):
        source_workdir = self.make_work_directory("source_workdir")
        target_workdir = self.make_work_directory("target_workdir")
        test_filename = "test_file"
        test_asset = {
            "source_location": source_workdir,
            "source_name": test_filename,
            "target_location": "",
            "target_name": test_filename,
            "copy_method": "copy",
        }
        self.make_empty_files(source_workdir, [test_filename])
        write_asset(test_asset, target_workdir)
        self.assertTrue(os.path.exists(os.path.join(target_workdir, test_filename)))


@pytest.mark.parametrize("target_location", ["", "subdir"])
def test_write_bytes_asset(tmpdir, target_location):
    asset = get_bytes_asset_dict(
        b"hello world", target_location=target_location, target_name="hello.txt"
    )
    write_asset(asset, tmpdir)
    dir_ = tmpdir.join(target_location)
    with dir_.join("hello.txt").open("rb") as f:
        assert f.read() == b"hello world"


def test_bytes_asset_serializes_with_yaml():
    asset = get_bytes_asset_dict(
        b"hello world", target_location="", target_name="hello.txt"
    )

    serialized = yaml.safe_dump(asset)
    deserialized = yaml.safe_load(serialized)

    assert deserialized == asset


def test_get_fv3config_yaml_asset(tmpdir):
    config = {"some": "dict", "diag_table": "/path/to/diag_table"}
    asset = get_fv3config_yaml_asset(config)
    write_asset(asset, str(tmpdir))

    with open(str(tmpdir.join("fv3config.yml"))) as f:
        loaded = yaml.safe_load(f)

    assert loaded == config


def test_get_diag_table_asset_from_class(tmpdir):
    field = fv3config.DiagFieldConfig("dycore", "u850", "u850")
    diag_table = fv3config.DiagTable(
        "experiment",
        datetime.datetime(2000, 1, 1),
        [fv3config.DiagFileConfig("name", 1, "hours", [field])],
    )
    asset = get_diag_table_asset({"diag_table": diag_table})
    write_asset(asset, str(tmpdir))

    with open(str(tmpdir.join("diag_table"))) as f:
        loaded = f.read()

    assert loaded == str(diag_table)


if __name__ == "__main__":
    unittest.main()



from collections import namedtuple
import pytest
import fv3config

MemoryArgs = namedtuple(
    "MemoryArgs", ["memory_gb", "memory_gb_limit_in", "memory_gb_limit"]
)


@pytest.fixture
def config_location():
    return "gs://my-bucket/fv3config.yml"


@pytest.fixture(params=["local", "remote"])
def outdir(request):
    return {"remote": "gs://my-bucket/rundir", "local": "/tmp/rundir"}[request.param]


@pytest.fixture(params=["default", "tagged"])
def docker_image(request):
    if request.param == "default":
        return "us.gcr.io/vcm-ml/fv3gfs-python"
    elif request.param == "tagged":
        return "us.gcr.io/vcm-ml/fv3gfs-python:1.0"


@pytest.fixture(params=["gcloud", "local"])
def runfile(request):
    if request.param == "gcloud":
        return "gs://my-bucket/runfile.py"
    if request.param == "local":
        return "/inside/docker/dir/runfile.py"


@pytest.fixture
def jobname():
    return "my-job-name"


@pytest.fixture(params=["default", "manual_limit"])
def memory_args(request):
    return {
        "default": MemoryArgs(3.6, None, 3.6),
        "manual_limit": MemoryArgs(3.6, 4.0, 4.0),
    }[request.param]


@pytest.fixture(params=["1_cpu", "2.5_cpu"])
def cpu_count(request):
    if request.param == "1_cpu":
        return 1
    elif request.param == "2.5_cpu":
        return 2.5


@pytest.fixture(params=["my-secret-name", None])
def gcp_secret(request):
    return request.param


@pytest.fixture(params=["Always", "IfNotPresent"])
def image_pull_policy(request):
    return request.param


@pytest.fixture(params=[{"job_group": "this_group", "extra_group": "this_extra"}, None])
def job_labels(request):
    return request.param


@pytest.mark.filterwarnings("ignore:Output directory .* is a local path.*")
def test_get_job(
    config_location,
    outdir,
    docker_image,
    runfile,
    jobname,
    memory_args,
    cpu_count,
    gcp_secret,
    image_pull_policy,
    job_labels,
):
    namespace = None
    job = fv3config.fv3run.run_kubernetes(
        config_location,
        outdir,
        docker_image,
        runfile,
        jobname,
        namespace,
        memory_args.memory_gb,
        memory_args.memory_gb_limit_in,
        cpu_count,
        gcp_secret,
        image_pull_policy,
        job_labels,
        submit=False,
    )
    _check_job(job, jobname)
    job_spec = job.spec
    _check_job_spec(job_spec)
    pod_spec = job_spec.template.spec
    assert pod_spec.restart_policy == "Never"
    container = pod_spec.containers[0]
    _check_secret(gcp_secret, container, pod_spec)
    assert container.image == docker_image
    assert container.image_pull_policy == image_pull_policy
    _check_resource_requirements(
        container, memory_args.memory_gb, memory_args.memory_gb_limit, cpu_count
    )
    assert len(pod_spec.tolerations) == 1
    toleration = pod_spec.tolerations[0]
    _check_toleration(toleration)
    _check_labels(job.spec.template.metadata, job_labels)


def _check_secret(gcp_secret, container, pod_spec):
    _check_secret_mount(gcp_secret, container, pod_spec)
    if gcp_secret is not None:
        _check_env(
            container.env,
            "GOOGLE_APPLICATION_CREDENTIALS",
            "/secret/gcp-credentials/key.json",
        )
        _check_env(
            container.env,
            "CLOUDSDK_AUTH_CREDENTIAL_FILE_OVERRIDE",
            "/secret/gcp-credentials/key.json",
        )
    else:
        _check_env(container.env, "GOOGLE_APPLICATION_CREDENTIALS", None)
        _check_env(container.env, "CLOUDSDK_AUTH_CREDENTIAL_FILE_OVERRIDE", None)


def _check_job(job, jobname):
    assert job.kind == "Job"
    assert job.api_version == "batch/v1"
    assert hasattr(job.metadata, "name")
    if jobname is not None:
        assert job.metadata.name == jobname


def _check_job_spec(job_spec):
    assert job_spec.backoff_limit == 0
    assert job_spec.completions == 1
    assert job_spec.ttl_seconds_after_finished == 100


def _check_toleration(toleration):
    assert toleration.effect == "NoSchedule"
    assert toleration.key == "dedicated"
    assert toleration.value == "climate-sim-pool"


def _check_secret_mount(gcp_secret, container, pod_spec):
    if gcp_secret is None:
        assert len(pod_spec.volumes) == 0
        assert len(container.volume_mounts) == 0
    else:
        assert len(pod_spec.volumes) == 1
        assert pod_spec.volumes[0].secret.secret_name == gcp_secret
        assert len(container.volume_mounts) == 1
        assert container.volume_mounts[0].name == pod_spec.volumes[0].name
        assert container.volume_mounts[0].read_only
        assert container.volume_mounts[0].mount_path == "/secret/gcp-credentials"


def _check_resource_requirements(container, memory_gb, memory_gb_limit, cpu_count):
    resource_reqs = container.resources
    assert resource_reqs.limits["memory"] == f"{memory_gb_limit:.1f}G"
    assert resource_reqs.requests["memory"] == f"{memory_gb:.1f}G"
    assert resource_reqs.requests["cpu"] == f"{cpu_count:.1f}"


def _check_env(env_list, key, value_or_none):
    if value_or_none is None:
        assert not any(env.name == key for env in env_list)
    else:
        assert any(env.name == key and env.value == value_or_none for env in env_list)


def _check_labels(metadata, job_labels):
    if job_labels is not None:
        for key, value in job_labels.items():
            assert key in metadata.labels
            assert metadata.labels[key] == value
    else:
        assert len(metadata.labels) == 1

    assert metadata.labels["app"] == "fv3run"



import os
import f90nml
from .._exceptions import InvalidFileError


def config_to_namelist(config, namelist_filename):
    """Write the namelist of a configuration dictionary to a namelist file.

    Args:
        config (dict): a configuration dictionary
        namelist_filename (str): filename to write, will be overwritten if present
    """
    if os.path.isfile(namelist_filename):
        os.remove(namelist_filename)
    f90nml.write(config["namelist"], namelist_filename)


def config_from_namelist(namelist_filename):
    """Read a configuration dictionary from a namelist file.

    Only reads the namelist configuration.

    Args:
        namelist_filename (str): a namelist filename

    Returns:
        return_dict (dict): a configuration dictionary

    Raises:
        InvalidFileError: if the specified filename does not exist
    """
    try:
        return_dict = _to_nested_dict(f90nml.read(namelist_filename).items())
    except FileNotFoundError:
        raise InvalidFileError(f"namelist {namelist_filename} does not exist")
    return return_dict


def _to_nested_dict(source):
    return_value = dict(source)
    for name, value in return_value.items():
        if isinstance(value, f90nml.Namelist):
            return_value[name] = _to_nested_dict(value)
    return return_value



import argparse
import fsspec

import fv3config


def _parse_write_run_directory_args():
    parser = argparse.ArgumentParser("write_run_directory")
    parser.add_argument(
        "config", help="URI to fv3config yaml file. Supports any path used by fsspec."
    )
    parser.add_argument(
        "rundir", help="Desired output directory. Must be a local directory"
    )
    return parser.parse_args()


def _parse_enable_restart_args():
    parser = argparse.ArgumentParser("enable_restart")
    parser.add_argument(
        "config",
        help="URI to fv3config yaml file. Supports any path used by fsspec. "
        "File will be modified in place.",
    )
    parser.add_argument(
        "initial_conditions", help="Path to restart initial conditions.",
    )
    return parser.parse_args()


def _parse_update_config_for_nudging_args():
    parser = argparse.ArgumentParser("update_config_for_nudging")
    parser.add_argument(
        "config",
        help="URI to fv3config yaml file. Supports any path used by fsspec. "
        "File will be modified in place.",
    )
    return parser.parse_args()


def write_run_directory():
    args = _parse_write_run_directory_args()

    with fsspec.open(args.config) as f:
        config = fv3config.load(f)

    fv3config.write_run_directory(config, args.rundir)


def enable_restart():
    args = _parse_enable_restart_args()

    with fsspec.open(args.config) as f:
        config = fv3config.load(f)

    restart_config = fv3config.enable_restart(config, args.initial_conditions)

    with fsspec.open(args.config, mode="w") as f:
        fv3config.dump(restart_config, f)


def update_config_for_nudging():
    args = _parse_update_config_for_nudging_args()

    with fsspec.open(args.config) as f:
        config = fv3config.load(f)

    # only update config if nudging is turned on
    if config["namelist"]["fv_core_nml"].get("nudge", False):
        fv3config.update_config_for_nudging(config)

        with fsspec.open(args.config, mode="w") as f:
            fv3config.dump(config, f)



from copy import deepcopy
import yaml
from typing import TextIO

from .types import Config
from .diag_table import DiagTable


def load(f: TextIO) -> Config:
    """Load a configuration from a file-like object f"""
    config = yaml.safe_load(f)
    if isinstance(config["diag_table"], dict):
        config["diag_table"] = DiagTable.from_dict(config["diag_table"])
    return config


def dump(config: Config, f: TextIO):
    """Serialize config to a file-like object using yaml encoding

    Args:
        config: an fv3config object
        f: the file like object to write to

    """
    config_copy = deepcopy(config)
    if isinstance(config["diag_table"], DiagTable):
        config_copy["diag_table"] = config["diag_table"].asdict()
    yaml.safe_dump(config_copy, f)



import collections
from datetime import datetime, timedelta
import os
from typing import Sequence, List, Mapping
import math
import fsspec
from .._asset_list import get_asset_dict
from .._exceptions import ConfigError
from ..filesystem import get_fs
from .derive import get_current_date, get_run_duration

SECONDS_IN_HOUR = 60 * 60


def _most_recent_nudge_time(start_time: datetime, interval: timedelta) -> datetime:
    """Return datetime object for the last nudging time preceding or concurrent
    with start_time, given interval between nudging times"""
    nudge_hours = range(0, 24, interval.seconds // SECONDS_IN_HOUR)
    first_nudge_hour = _most_recent_hour(start_time.hour, nudge_hours)
    return datetime(start_time.year, start_time.month, start_time.day, first_nudge_hour)


def _most_recent_hour(current_hour: int, hour_array: Sequence[int]) -> int:
    """Return latest hour in hour_array that precedes or is concurrent with
    current_hour"""
    for hour in hour_array:
        if hour <= current_hour:
            first_nudge_hour = hour
    return first_nudge_hour


def _get_nudge_time_list(
    run_duration: timedelta, current_date: Sequence[int], interval: timedelta
) -> List[datetime]:
    """Return list of datetime objects corresponding to times at which analysis files
    are required for nudging for a given model run configuration"""
    start_time = datetime(*current_date)
    first_nudge_time = _most_recent_nudge_time(start_time, interval)
    nudge_duration = run_duration + (start_time - first_nudge_time)
    nudge_duration_hours = int(
        math.ceil(nudge_duration.total_seconds() / SECONDS_IN_HOUR)
    )
    interval_hours = interval.seconds // SECONDS_IN_HOUR
    nudging_hours = range(0, nudge_duration_hours + interval_hours, interval_hours)
    return [first_nudge_time + timedelta(hours=hour) for hour in nudging_hours]


def get_nudging_assets(
    run_duration: timedelta,
    current_date: Sequence[int],
    nudge_path: str,
    nudge_filename_pattern: str = "%Y%m%d_%HZ_T85LR.nc",
    copy_method: str = "copy",
    nudge_interval: timedelta = timedelta(hours=6),
) -> List[Mapping]:
    """Return list of assets of nudging files required for given run duration and
    start time.
    
    This method defines file paths directly from its arguments, without
    determining whether the files themselves are present.
    
    Args:
        run_duration: length of fv3gfs run
        current_date: start time of fv3gfs run as a sequence of 6 integers
        nudge_path: local or remote path to nudging files
        nudge_filename_pattern: template for nudging filenames. Pattern should follow
            style of datetime strptime and strftime 'format' argument. Defaults to
            '%Y%m%d_%HZ_T85LR.nc'.
        copy_method: copy_method for nudging file assets. Defaults to 'copy'.
        nudge_interval: time between nudging files. Must be multiple of 1 hour.
            Defaults to 6 hours.

    Returns:
        list of all assets required for nudging run

    Raises:
        ConfigError: if copy_method is "link" and a remote path is given for nudge_path
    """
    if get_fs(nudge_path) != fsspec.filesystem("file") and copy_method == "link":
        raise ConfigError(
            "Cannot link nudging files if using remote path for nudge_path. "
            f"Got {nudge_path}."
        )
    time_list = _get_nudge_time_list(run_duration, current_date, nudge_interval)
    filename_list = [time.strftime(nudge_filename_pattern) for time in time_list]
    nudging_assets = [
        get_asset_dict(
            nudge_path, file_, target_location="INPUT", copy_method=copy_method
        )
        for file_ in filename_list
    ]
    return nudging_assets


def _non_nudging_assets(
    assets: Sequence[Mapping], filename_pattern: str,
) -> List[Mapping]:
    """Given list of assets, return filtered list with no nudging assets."""
    return [item for item in assets if not _is_nudging_asset(item, filename_pattern)]


def _is_nudging_asset(item, pattern):
    if not isinstance(item, collections.abc.Mapping):
        # not an asset
        return False
    try:
        target_name = item["target_name"]
        datetime.strptime(target_name, pattern)
        return True
    except (KeyError, ValueError):
        # not an asset or target_name does not fit given pattern
        return False


def _clear_nudging_assets(config):
    """Remove any assets in patch_files that match given filename_pattern"""
    pattern = config["gfs_analysis_data"]["filename_pattern"]
    if "patch_files" in config:
        config["patch_files"] = _non_nudging_assets(config["patch_files"], pattern)


def update_config_for_nudging(config: Mapping):
    """Update config object in place to include nudging file assets and associated
    file_names namelist entry. Requires 'gfs_analysis_data' entry in fv3config object
    with 'url' and 'filename_pattern' entries.
    
    Args:
        config: configuration dictionary

    Raises:
        ConfigError: if provided config does not contain "gfs_analysis_data" section.

    Note:
        will delete any existing assets in 'patch_files' that match the given
        filename_pattern before new assets are added.
    """
    gfs_analysis_data = config.get("gfs_analysis_data", {})
    if "url" not in gfs_analysis_data or "filename_pattern" not in gfs_analysis_data:
        raise ConfigError(
            "Config must contain 'gfs_analysis_data' section with 'url' and"
            "'filename_pattern' items if 'namelist.fv_core_nml.nudge' is True."
        )

    _clear_nudging_assets(config)

    nudging_file_assets = get_nudging_assets(
        get_run_duration(config),
        get_current_date(config),
        gfs_analysis_data["url"],
        nudge_filename_pattern=gfs_analysis_data["filename_pattern"],
        copy_method=gfs_analysis_data.get("copy_method", "copy"),
        nudge_interval=gfs_analysis_data.get("interval", timedelta(hours=6)),
    )

    target_file_paths = [
        os.path.join(asset["target_location"], asset["target_name"])
        for asset in nudging_file_assets
    ]

    namelist = config["namelist"]
    namelist.setdefault("fv_nwp_nudge_nml", {})["file_names"] = target_file_paths
    config.setdefault("patch_files", []).extend(nudging_file_assets)



def c12_config():
    return {
        "data_table": "default",
        "diag_table": "default",
        "experiment_name": "default",
        "forcing": "gs://vcm-fv3config/data/base_forcing/v1.1/",
        "orographic_forcing": "gs://vcm-fv3config/data/orographic_data/v1.0",
        "initial_conditions": "gs://vcm-fv3config/data/initial_conditions/gfs_c12_example/v1.0",
        "namelist": {
            "amip_interp_nml": {
                "data_set": "reynolds_oi",
                "date_out_of_range": "climo",
                "interp_oi_sst": True,
                "no_anom_sst": False,
                "use_ncep_ice": False,
                "use_ncep_sst": True,
            },
            "atmos_model_nml": {
                "blocksize": 24,
                "chksum_debug": False,
                "dycore_only": False,
                "fdiag": 0.0,
                "fhmax": 1024.0,
                "fhmaxhf": -1.0,
                "fhout": 0.25,
                "fhouthf": 0.0,
            },
            "cires_ugwp_nml": {
                "knob_ugwp_azdir": [2, 4, 4, 4],
                "knob_ugwp_doaxyz": 1,
                "knob_ugwp_doheat": 1,
                "knob_ugwp_dokdis": 0,
                "knob_ugwp_effac": [1, 1, 1, 1],
                "knob_ugwp_ndx4lh": 4,
                "knob_ugwp_solver": 2,
                "knob_ugwp_source": [1, 1, 1, 0],
                "knob_ugwp_stoch": [0, 0, 0, 0],
                "knob_ugwp_version": 0,
                "knob_ugwp_wvspec": [1, 32, 32, 32],
                "launch_level": 55,
            },
            "coupler_nml": {
                "atmos_nthreads": 1,
                "calendar": "julian",
                "current_date": [2016, 8, 1, 0, 0, 0],
                "days": 0,
                "dt_atmos": 900,
                "dt_ocean": 900,
                "hours": 0,
                "memuse_verbose": True,
                "minutes": 30,
                "months": 0,
                "ncores_per_node": 32,
                "seconds": 0,
                "use_hyper_thread": True,
            },
            "diag_manager_nml": {"prepend_date": False},
            "external_ic_nml": {
                "checker_tr": False,
                "filtered_terrain": True,
                "gfs_dwinds": True,
                "levp": 64,
                "nt_checker": 0,
            },
            "fms_io_nml": {
                "checksum_required": False,
                "max_files_r": 100,
                "max_files_w": 100,
            },
            "fms_nml": {
                "clock_grain": "ROUTINE",
                "domains_stack_size": 3000000,
                "print_memory_usage": False,
            },
            "fv_core_nml": {
                "a_imp": 1.0,
                "adjust_dry_mass": False,
                "beta": 0.0,
                "consv_am": False,
                "consv_te": 1.0,
                "d2_bg": 0.0,
                "d2_bg_k1": 0.16,
                "d2_bg_k2": 0.02,
                "d4_bg": 0.15,
                "d_con": 1.0,
                "d_ext": 0.0,
                "dddmp": 0.2,
                "delt_max": 0.002,
                "dnats": 1,
                "do_sat_adj": True,
                "do_vort_damp": True,
                "dwind_2d": False,
                "external_ic": True,
                "fill": True,
                "fv_debug": False,
                "fv_sg_adj": 900,
                "gfs_phil": False,
                "hord_dp": 6,
                "hord_mt": 6,
                "hord_tm": 6,
                "hord_tr": 8,
                "hord_vt": 6,
                "hydrostatic": False,
                "io_layout": [1, 1],
                "k_split": 1,
                "ke_bg": 0.0,
                "kord_mt": 10,
                "kord_tm": -10,
                "kord_tr": 10,
                "kord_wz": 10,
                "layout": [1, 1],
                "make_nh": True,
                "mountain": False,
                "n_split": 6,
                "n_sponge": 4,
                "na_init": 1,
                "ncep_ic": False,
                "nggps_ic": True,
                "no_dycore": False,
                "nord": 2,
                "npx": 13,
                "npy": 13,
                "npz": 63,
                "ntiles": 6,
                "nudge": False,
                "nudge_qv": True,
                "nwat": 6,
                "p_fac": 0.1,
                "phys_hydrostatic": False,
                "print_freq": 3,
                "range_warn": True,
                "reset_eta": False,
                "rf_cutoff": 800.0,
                "rf_fast": False,
                "tau": 5.0,
                "use_hydro_pressure": False,
                "vtdm4": 0.06,
                "warm_start": False,
                "z_tracer": True,
            },
            "fv_grid_nml": {},
            "gfdl_cloud_microphysics_nml": {
                "c_cracw": 0.8,
                "c_paut": 0.5,
                "c_pgacs": 0.01,
                "c_psaci": 0.05,
                "ccn_l": 300.0,
                "ccn_o": 100.0,
                "const_vg": False,
                "const_vi": False,
                "const_vr": False,
                "const_vs": False,
                "de_ice": False,
                "do_qa": True,
                "do_sedi_heat": False,
                "dw_land": 0.16,
                "dw_ocean": 0.1,
                "fast_sat_adj": True,
                "fix_negative": True,
                "icloud_f": 1,
                "mono_prof": True,
                "mp_time": 450.0,
                "prog_ccn": False,
                "qi0_crt": 8e-05,
                "qi_lim": 1.0,
                "ql_gen": 0.001,
                "ql_mlt": 0.001,
                "qs0_crt": 0.001,
                "rad_graupel": True,
                "rad_rain": True,
                "rad_snow": True,
                "rh_inc": 0.3,
                "rh_inr": 0.3,
                "rh_ins": 0.3,
                "rthresh": 1e-05,
                "sedi_transport": False,
                "tau_g2v": 900.0,
                "tau_i2s": 1000.0,
                "tau_l2v": [225.0],
                "tau_v2l": 150.0,
                "use_ccn": True,
                "use_ppm": False,
                "vg_max": 12.0,
                "vi_max": 1.0,
                "vr_max": 12.0,
                "vs_max": 2.0,
                "z_slope_ice": True,
                "z_slope_liq": True,
            },
            "gfs_physics_nml": {
                "cal_pre": False,
                "cdmbgwd": [3.5, 0.25],
                "cnvcld": False,
                "cnvgwd": True,
                "debug": False,
                "dspheat": True,
                "fhcyc": 24.0,
                "fhlwr": 3600.0,
                "fhswr": 3600.0,
                "fhzero": 0.25,
                "hybedmf": True,
                "iaer": 111,
                "ialb": 1,
                "ico2": 2,
                "iems": 1,
                "imfdeepcnv": 2,
                "imfshalcnv": 2,
                "imp_physics": 11,
                "isol": 2,
                "isot": 1,
                "isubc_lw": 2,
                "isubc_sw": 2,
                "ivegsrc": 1,
                "ldiag3d": False,
                "lwhtr": True,
                "ncld": 5,
                "nst_anl": True,
                "pdfcld": False,
                "pre_rad": False,
                "prslrd0": 0.0,
                "random_clds": False,
                "redrag": True,
                "shal_cnv": True,
                "swhtr": True,
                "trans_trac": True,
                "use_ufo": True,
            },
            "interpolator_nml": {"interp_method": "conserve_great_circle"},
            "nam_stochy": {"lat_s": 96, "lon_s": 192, "ntrunc": 94},
            "namsfc": {
                "fabsl": 99999,
                "faisl": 99999,
                "faiss": 99999,
                "fnabsc": "grb/global_mxsnoalb.uariz.t1534.3072.1536.rg.grb",
                "fnacna": "",
                "fnaisc": "grb/CFSR.SEAICE.1982.2012.monthly.clim.grb",
                "fnalbc": "grb/global_snowfree_albedo.bosu.t1534.3072.1536.rg.grb",
                "fnalbc2": "grb/global_albedo4.1x1.grb",
                "fnglac": "grb/global_glacier.2x2.grb",
                "fnmskh": "grb/seaice_newland.grb",
                "fnmxic": "grb/global_maxice.2x2.grb",
                "fnslpc": "grb/global_slope.1x1.grb",
                "fnsmcc": "grb/global_soilmgldas.t1534.3072.1536.grb",
                "fnsnoa": "",
                "fnsnoc": "grb/global_snoclim.1.875.grb",
                "fnsotc": "grb/global_soiltype.statsgo.t1534.3072.1536.rg.grb",
                "fntg3c": "grb/global_tg3clim.2.6x1.5.grb",
                "fntsfa": "",
                "fntsfc": "grb/RTGSST.1982.2012.monthly.clim.grb",
                "fnvegc": "grb/global_vegfrac.0.144.decpercent.grb",
                "fnvetc": "grb/global_vegtype.igbp.t1534.3072.1536.rg.grb",
                "fnvmnc": "grb/global_shdmin.0.144x0.144.grb",
                "fnvmxc": "grb/global_shdmax.0.144x0.144.grb",
                "fnzorc": "igbp",
                "fsicl": 99999,
                "fsics": 99999,
                "fslpl": 99999,
                "fsmcl": [99999, 99999, 99999],
                "fsnol": 99999,
                "fsnos": 99999,
                "fsotl": 99999,
                "ftsfl": 99999,
                "ftsfs": 90,
                "fvetl": 99999,
                "fvmnl": 99999,
                "fvmxl": 99999,
                "ldebug": False,
            },
        },
    }



import logging
import os
from .namelist import config_to_namelist
from .._asset_list import write_assets_to_directory
from .._tables import update_diag_table_for_config
from .derive import get_current_date
from .nudging import update_config_for_nudging

logger = logging.getLogger("fv3config")


def write_run_directory(config, target_directory):
    """Write a run directory based on a configuration dictionary.

    Args:
        config (dict): a configuration dictionary
        target_directory (str): target directory, will be created if it does not exist
    """
    logger.debug(f"Writing run directory to {target_directory}")
    if config["namelist"]["fv_core_nml"].get("nudge", False):
        update_config_for_nudging(config)
    write_assets_to_directory(config, target_directory)
    os.makedirs(os.path.join(target_directory, "RESTART"), exist_ok=True)
    current_date = get_current_date(config)
    update_diag_table_for_config(
        config, current_date, os.path.join(target_directory, "diag_table")
    )
    config_to_namelist(config, os.path.join(target_directory, "input.nml"))



import collections
import contextlib
import os
import shutil
import subprocess
import sys
import tempfile
import unittest
import unittest.mock

import gcsfs
import pytest

import fv3config
from fv3config.fv3run._native import (
    RUNFILE_ENV_VAR,
    _output_stream_context,
    _get_python_command,
    call_via_subprocess,
)

TEST_DIR = os.path.dirname(os.path.realpath(__file__))
MOCK_RUNSCRIPT = os.path.abspath(os.path.join(TEST_DIR, "testdata/mock_runscript.py"))
DOCKER_IMAGE_NAME = "us.gcr.io/vcm-ml/fv3gfs-python"
USER_UID_GID = f"{os.getuid()}:{os.getgid()}"

MockTempfile = collections.namedtuple("MockTempfile", ["name"])

MPI_ENABLED = shutil.which("mpirun") is not None
try:

    subprocess.check_call(
        ["docker", "image", "inspect", DOCKER_IMAGE_NAME],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    DOCKER_ENABLED = True
except (subprocess.CalledProcessError, FileNotFoundError):
    DOCKER_ENABLED = False


def subprocess_run(config_dict, outdir):
    with tempfile.NamedTemporaryFile(mode="w") as config_file:
        fv3config.dump(config_dict, config_file)
        config_file.flush()
        subprocess.check_call(
            [
                "fv3run",
                config_file.name,
                outdir,
                "--runfile",
                MOCK_RUNSCRIPT,
                "--capture-output",
            ]
        )


def docker_run(config_dict, outdir):
    with tempfile.NamedTemporaryFile(mode="w") as config_file:
        fv3config.dump(config_dict, config_file)
        config_file.flush()
        try:
            subprocess.check_call(
                [
                    "fv3run",
                    config_file.name,
                    outdir,
                    "--runfile",
                    MOCK_RUNSCRIPT,
                    "--dockerimage",
                    DOCKER_IMAGE_NAME,
                ]
            )
        except subprocess.CalledProcessError as err:
            if os.path.isfile(os.path.join(outdir, "stdout.log")):
                with open(os.path.join(outdir, "stdout.log")) as stdout:
                    print(f"STDOUT: {stdout.read()}")
            if os.path.isfile(os.path.join(outdir, "stderr.log")):
                with open(os.path.join(outdir, "stderr.log")) as stderr:
                    print(f"STDERR: {stderr.read()}")
            raise err


def config_dict_module_run(config_dict, outdir):
    fv3config.run_native(config_dict, outdir, runfile=MOCK_RUNSCRIPT)


def config_dict_filename_run(config_dict, outdir):
    with tempfile.NamedTemporaryFile(mode="w") as config_file:
        fv3config.dump(config_dict, config_file)
        config_file.flush()
        fv3config.run_native(config_file.name, outdir, runfile=MOCK_RUNSCRIPT)


def count_executed_ranks(rundir):
    filenames = os.listdir(rundir)
    total = 0
    while f"rank_{total}" in filenames:
        total += 1
    return total


@pytest.fixture
def config(c12_config):
    return c12_config


@contextlib.contextmanager
def cleaned_up_directory(dirname):
    try:
        yield
    finally:
        if os.path.isdir(dirname):
            shutil.rmtree(dirname)


def check_run_directory(dirname):
    filenames = os.listdir(dirname)
    assert "stdout.log" in filenames
    assert "stderr.log" in filenames
    assert count_executed_ranks(dirname) == 6
    assert "input.nml" in filenames
    assert "INPUT" in filenames


@pytest.mark.filterwarnings("ignore:could not remove stacksize limit")
@pytest.mark.parametrize("runner", [config_dict_module_run, config_dict_filename_run])
def test_fv3run_with_mocked_subprocess(runner, config):
    outdir = os.path.join(TEST_DIR, "outdir")

    with unittest.mock.patch("subprocess.check_call") as mock, cleaned_up_directory(
        outdir
    ):
        runner(config, outdir)
        assert mock.called
        call_args = list(mock.call_args[0])
        # ensure test does not depend on # of processors on testing system
        while "--oversubscribe" in call_args[0]:
            call_args[0].remove("--oversubscribe")
        assert call_args == [
            [
                "mpirun",
                "-n",
                "6",
                "--allow-run-as-root",
                "--use-hwthread-cpus",
                "--mca",
                "btl_vader_single_copy_mechanism",
                "none",
                "python3",
                "-m",
                "mpi4py",
                "mock_runscript.py",
            ]
        ]
        with open(os.path.join(outdir, "fv3config.yml"), "r") as f:
            written_config = fv3config.load(f)
        assert written_config == config


@pytest.mark.parametrize(
    "runfile, expected_bind_mount_args, expected_python_args",
    [
        ["/tmp/runfile.py", ["-v", "/tmp/runfile.py:/runfile.py"], "/runfile.py"],
        [
            "relative.py",
            ["-v", f"{os.path.join(os.getcwd(), 'relative.py')}:/runfile.py"],
            "/runfile.py",
        ],
        ["gs://bucket-name/runfile.py", [], "gs://bucket-name/runfile.py"],
    ],
)
def test_get_runfile_args(runfile, expected_bind_mount_args, expected_python_args):
    bind_mount_args = []
    runfile = fv3config.fv3run._docker._get_runfile_args(runfile, bind_mount_args)
    assert bind_mount_args == expected_bind_mount_args


@pytest.mark.parametrize(
    "runfile, env_var, expected",
    [
        ("/path/to/runfile.py", None, ["python3", "-m", "mpi4py", "runfile.py"]),
        ("a", "b", ["python3", "-m", "mpi4py", "a"]),
        (
            None,
            "/path/to/runfile.py",
            ["python3", "-m", "mpi4py", "/path/to/runfile.py"],
        ),
        (None, None, ["python3", "-m", "mpi4py", "-m", "fv3gfs.run"]),
    ],
)
def test__get_native_python_args(monkeypatch, runfile, expected, env_var):
    if env_var is not None:
        monkeypatch.setenv(RUNFILE_ENV_VAR, env_var)
    assert _get_python_command(runfile) == expected


_original_get_file = fv3config.filesystem.get_file


def test__output_stream_context_captured(tmpdir):
    err_msg = b"error"
    out_msg = b"output"
    with _output_stream_context(tmpdir, True) as (stdout, stderr):
        stdout.write(out_msg)
        stderr.write(err_msg)

    with open(tmpdir.join("stderr.log"), "rb") as f:
        assert err_msg == f.read()

    with open(tmpdir.join("stdout.log"), "rb") as f:
        assert out_msg == f.read()


def test__output_stream_context_uncaptured(tmpdir):
    with _output_stream_context(tmpdir, False) as (stdout, stderr):
        assert stdout == sys.stdout
        assert stderr == sys.stderr


def maybe_get_file(*args, **kwargs):
    try:
        _original_get_file(*args, **kwargs)
    except (OSError, gcsfs.utils.HttpError):
        pass


@pytest.mark.parametrize(
    "outdir, expected_docker_args, expected_bind_mount_args",
    [
        [
            "relative/",
            ["--rm", "--user", USER_UID_GID],
            ["-v", f"{os.path.join(os.getcwd(), 'relative')}:/outdir"],
        ],
        ["/abs/path/", ["--rm", "--user", USER_UID_GID], ["-v", f"/abs/path:/outdir"]],
    ],
)
def test_get_docker_args(outdir, expected_docker_args, expected_bind_mount_args):
    docker_args = []
    bind_mount_args = []
    fv3config.fv3run._docker._get_docker_args(docker_args, bind_mount_args, outdir)
    assert docker_args == expected_docker_args
    assert bind_mount_args == expected_bind_mount_args


@pytest.mark.parametrize(
    "keyfile, expected_docker_args, expected_bind_mount_args",
    [
        [None, [], []],
        [
            "relative_key.json",
            ["-e", "GOOGLE_APPLICATION_CREDENTIALS=/gcs_key.json"],
            ["-v", f"{os.path.join(os.getcwd(), 'relative_key.json')}:/gcs_key.json"],
        ],
        [
            "/abs/key.json",
            ["-e", "GOOGLE_APPLICATION_CREDENTIALS=/gcs_key.json"],
            ["-v", "/abs/key.json:/gcs_key.json"],
        ],
    ],
)
def test_get_credentials_args(keyfile, expected_docker_args, expected_bind_mount_args):
    docker_args = []
    bind_mount_args = []
    fv3config.fv3run._docker._get_credentials_args(
        keyfile, docker_args, bind_mount_args
    )
    assert docker_args == expected_docker_args
    assert bind_mount_args == expected_bind_mount_args


@pytest.mark.parametrize(
    "config_dict, local_paths",
    [
        [
            {
                "experiment_name": "default_experiment",
                "initial_conditions": "default",
                "forcing": "gs://bucket/forcing/default",
                "diag_table": "gs://bucket/diag/default",
                "data_table": "gs://bucket/data_table/default",
            },
            [],
        ],
        [
            {
                "experiment_name": "default_experiment",
                "initial_conditions": "/local/initial/conditions",
                "forcing": "gs://bucket/forcing/default",
                "diag_table": "gs://bucket/diag/default",
                "data_table": "gs://bucket/data_table/default",
            },
            ["/local/initial/conditions"],
        ],
        [
            {
                "experiment_name": "default_experiment",
                "initial_conditions": "/local/initial/conditions",
                "forcing": "/local/forcing/default",
                "diag_table": "/local/diag/default",
                "data_table": "gs://bucket/data_table/default",
            },
            [
                "/local/initial/conditions",
                "/local/forcing/default",
                "/local/diag/default",
            ],
        ],
        [
            {
                "experiment_name": "default_experiment",
                "initial_conditions": "gs://bucket/initial_conditions/default",
                "forcing": [
                    {
                        "source_location": "/local/directory",
                        "source_name": "source_name.nc",
                        "target_location": "INPUT/",
                        "target_name": "filename.nc",
                        "copy_method": "copy",
                    }
                ],
                "diag_table": "/local/diag/default",
                "data_table": "gs://bucket/data_table/default",
            },
            ["/local/directory/source_name.nc", "/local/diag/default"],
        ],
        [
            {
                "experiment_name": "default_experiment",
                "initial_conditions": "gs://bucket/initial_conditions/default",
                "forcing": "gs://bucket/forcing/default",
                "diag_table": [
                    {
                        "source_location": "gs://remote/directory",
                        "source_name": "source_name.nc",
                        "target_location": "INPUT/",
                        "target_name": "filename.nc",
                        "copy_method": "copy",
                    }
                ],
                "data_table": "gs://bucket/data_table/default",
            },
            [],
        ],
        [
            {
                "experiment_name": "default_experiment",
                "initial_conditions": "gs://bucket/initial_conditions/default",
                "forcing": "gs://bucket/forcing/default",
                "diag_table": [
                    {
                        "source_location": "gs://remote/directory",
                        "source_name": "source_name.nc",
                        "target_location": "INPUT/",
                        "target_name": "filename.nc",
                        "copy_method": "copy",
                    }
                ],
                "data_table": "gs://bucket/data_table/default",
                "patch_files": [],
            },
            [],
        ],
        [
            {
                "experiment_name": "default_experiment",
                "initial_conditions": "gs://bucket/initial_conditions/default",
                "forcing": "gs://bucket/forcing/default",
                "diag_table": [
                    {
                        "source_location": "gs://remote/directory",
                        "source_name": "source_name.nc",
                        "target_location": "INPUT/",
                        "target_name": "filename.nc",
                        "copy_method": "copy",
                    }
                ],
                "data_table": "gs://bucket/data_table/default",
                "patch_files": {
                    "source_location": "gs://remote/directory",
                    "source_name": "source_name.nc",
                    "target_location": "INPUT/",
                    "target_name": "filename.nc",
                    "copy_method": "copy",
                },
            },
            [],
        ],
        [
            {
                "experiment_name": "default_experiment",
                "initial_conditions": "gs://bucket/initial_conditions/default",
                "forcing": "gs://bucket/forcing/default",
                "diag_table": [
                    {
                        "source_location": "gs://remote/directory",
                        "source_name": "source_name.nc",
                        "target_location": "INPUT/",
                        "target_name": "filename.nc",
                        "copy_method": "copy",
                    }
                ],
                "data_table": "gs://bucket/data_table/default",
                "patch_files": [
                    {
                        "source_location": "/local/directory",
                        "source_name": "source_name.nc",
                        "target_location": "INPUT/",
                        "target_name": "filename.nc",
                        "copy_method": "copy",
                    }
                ],
            },
            ["/local/directory/source_name.nc"],
        ],
    ],
)
def test_get_local_paths(config_dict, local_paths):
    return_value = fv3config.fv3run._docker._get_local_data_paths(config_dict)
    assert set(return_value) == set(local_paths)  # order does not matter


def test_call_via_subprocess_command():
    import json

    @call_via_subprocess(__name__)
    def dummy_function(*obj, **kwargs):
        pass

    command = dummy_function.command(1, a=1, b=1)
    args = command.pop()
    assert json.loads(args) == [[1], dict(a=1, b=1)]
    assert command == ["python", "-m", dummy_function.__module__]


def test_call_via_subprocess_main():

    # need to append outputs here since call_via_subprocess does not
    # support return arguments
    output = []

    @call_via_subprocess(__name__)
    def dummy_function(*obj):
        output.append(hash(obj))

    inputs = (1, 2, 3)

    # first call is via "main"
    serialized_args = dummy_function.command(*inputs)[-1]
    argv = [None, serialized_args]
    dummy_function.main(argv)
    dummy_ans = output.pop()

    # second call is normal python call
    dummy_function(*inputs)
    python_ans = output.pop()

    # assert results where the same
    assert dummy_ans == python_ans


def test_call_via_subprocess_command_fails_with_bad_args():
    @call_via_subprocess(__name__)
    def dummy_function(a, b):
        pass

    with pytest.raises(TypeError):
        dummy_function.command(1, 2, 3, king="kong")



import unittest
import tempfile
import pytest
import fv3config
import os
import shutil
import copy
import datetime

from .mocks import c12_config

TEST_DIR = os.path.dirname(os.path.abspath(__file__))


DEFAULT_CONFIG = c12_config()


@pytest.mark.parametrize(
    "source_filename, cache_subpath",
    [
        ("gs://my-bucket/my_filename.txt", "gs/my-bucket/my_filename.txt"),
        ("http://www.mysite.com/dir/file.nc", "http/www.mysite.com/dir/file.nc"),
    ],
)
def test_cache_filename(source_filename, cache_subpath):
    cache_dir = fv3config.caching.get_internal_cache_dir()
    cache_filename = os.path.join(cache_dir, cache_subpath)
    result = fv3config.filesystem._get_cache_filename(source_filename)
    assert result == cache_filename


def test_cache_filename_raises_on_no_filename():
    with pytest.raises(ValueError):
        fv3config.filesystem._get_cache_filename("gs://")


class ChangeCacheDirectoryTests(unittest.TestCase):
    def test_get_archive_dir(self):
        result = fv3config.caching.get_cache_dir()
        assert isinstance(result, str)
        assert os.path.isabs(result)

    def test_set_then_get_archive_dir(self):
        with tempfile.TemporaryDirectory() as tempdir:
            original = fv3config.caching.get_cache_dir()
            try:
                fv3config.caching.set_cache_dir(tempdir)
                new = fv3config.caching.get_cache_dir()
                assert new != original
                assert new == tempdir
            finally:
                fv3config.caching.set_cache_dir(original)


class CacheDirectoryTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cache_dir = tempfile.TemporaryDirectory()
        cls.original_cache_dir = fv3config.caching.get_cache_dir()
        fv3config.caching.set_cache_dir(cls.cache_dir.name)

    @classmethod
    def tearDownClass(cls):
        fv3config.caching.set_cache_dir(cls.original_cache_dir)
        cls.cache_dir.cleanup()

    def tearDown(self):
        gs_cache_dir = os.path.join(self.cache_dir.name, "fv3config-cache/gs/")
        if os.path.isdir(gs_cache_dir):
            shutil.rmtree(gs_cache_dir)

    def test_cache_diag_table(self):
        config = copy.deepcopy(DEFAULT_CONFIG)
        config[
            "diag_table"
        ] = "gs://vcm-fv3config/config/diag_table/default/v1.0/diag_table"
        cache_filename = os.path.join(
            fv3config.get_cache_dir(),
            "fv3config-cache/gs/vcm-fv3config/config/diag_table/default/v1.0/diag_table",
        )
        assert not os.path.isfile(cache_filename)
        with tempfile.TemporaryDirectory() as rundir:
            fv3config.write_run_directory(config, rundir)
        assert os.path.isfile(cache_filename)

    def test_disable_caching(self):
        config = copy.deepcopy(DEFAULT_CONFIG)
        config[
            "diag_table"
        ] = "gs://vcm-fv3config/config/diag_table/default/v1.0/diag_table"
        cache_filename = os.path.join(
            fv3config.get_cache_dir(),
            "fv3config-cache/gs/vcm-fv3config/config/diag_table/default/v1.0/diag_table",
        )
        assert not os.path.isfile(cache_filename)
        fv3config.do_remote_caching(False)
        with tempfile.TemporaryDirectory() as rundir:
            fv3config.write_run_directory(config, rundir)
        assert not os.path.isfile(cache_filename)

    def test_reenable_caching(self):
        config = copy.deepcopy(DEFAULT_CONFIG)
        config[
            "diag_table"
        ] = "gs://vcm-fv3config/config/diag_table/default/v1.0/diag_table"
        cache_filename = os.path.join(
            fv3config.get_cache_dir(),
            "fv3config-cache/gs/vcm-fv3config/config/diag_table/default/v1.0/diag_table",
        )
        assert not os.path.isfile(cache_filename)
        fv3config.do_remote_caching(False)
        fv3config.do_remote_caching(True)
        with tempfile.TemporaryDirectory() as rundir:
            fv3config.write_run_directory(config, rundir)
        assert os.path.isfile(cache_filename)

    def test_cached_diag_table_is_not_redownloaded(self):
        config = copy.deepcopy(DEFAULT_CONFIG)
        config[
            "diag_table"
        ] = "gs://vcm-fv3config/config/diag_table/default/v1.0/diag_table"
        with tempfile.TemporaryDirectory() as rundir:
            fv3config.write_run_directory(config, rundir)
        cache_filename = os.path.join(
            fv3config.get_cache_dir(),
            "fv3config-cache/gs/vcm-fv3config/config/diag_table/default/v1.0/diag_table",
        )
        assert os.path.isfile(cache_filename)
        modification_time = os.path.getmtime(cache_filename)
        with tempfile.TemporaryDirectory() as rundir:
            fv3config.write_run_directory(config, rundir)
        assert os.path.getmtime(cache_filename) == modification_time

    def test_rundir_contains_fv3config_yml(self):
        config = copy.deepcopy(DEFAULT_CONFIG)
        with tempfile.TemporaryDirectory() as rundir:
            fv3config.write_run_directory(config, rundir)
            assert "fv3config.yml" in os.listdir(rundir)

    def test_write_run_directory_succeeds_with_diag_table_class(self):
        config = copy.deepcopy(DEFAULT_CONFIG)
        start_time = datetime.datetime(2000, 1, 1)
        config["diag_table"] = fv3config.DiagTable("name", start_time, [])
        with tempfile.TemporaryDirectory() as rundir:
            fv3config.write_run_directory(config, rundir)

    def test_rundir_contains_nudging_asset_if_enabled(self):
        config = copy.deepcopy(DEFAULT_CONFIG)
        config["gfs_analysis_data"] = {
            "url": "gs://vcm-fv3config/data/gfs_nudging_data/v1.0",
            "filename_pattern": "%Y%m%d_%H.nc",
        }
        config["namelist"]["fv_core_nml"]["nudge"] = True
        with tempfile.TemporaryDirectory() as rundir:
            fv3config.write_run_directory(config, rundir)
            assert "20160801_00.nc" in os.listdir(os.path.join(rundir, "INPUT"))
            assert "20160801_06.nc" in os.listdir(os.path.join(rundir, "INPUT"))



from datetime import datetime
import pytest

from fv3config.config.diag_table import (
    DiagTable,
    DiagFieldConfig,
    DiagFileConfig,
)
from fv3config._exceptions import ConfigError


def diag_field():
    return DiagFieldConfig("dynamics", "u850", "zonal_wind_at_850hPa")


def diag_file(name):
    return DiagFileConfig(name, 1, "hours", [diag_field(), diag_field()])


@pytest.fixture
def diag_table():
    return DiagTable(
        "experiment",
        datetime(2000, 1, 1),
        [diag_file("first_diagnostics"), diag_file("second_diagnostics")],
    )


def test_DiagTable_string_round_trip(diag_table):
    round_tripped_diag_table = DiagTable.from_str(str(diag_table))
    assert str(diag_table) == str(round_tripped_diag_table)


def test_DiagTable_dict_round_trip(diag_table):
    round_tripped_diag_table = DiagTable.from_dict(diag_table.asdict())
    assert str(diag_table) == str(round_tripped_diag_table)


def test_DiagTable__repr__(diag_table):
    expected_str = """experiment
2000 1 1 0 0 0

"first_diagnostics", 1, "hours", 1, "hours", "time"
"second_diagnostics", 1, "hours", 1, "hours", "time"

"dynamics", "u850", "zonal_wind_at_850hPa", "first_diagnostics", "all", "none", "none", 2
"dynamics", "u850", "zonal_wind_at_850hPa", "first_diagnostics", "all", "none", "none", 2

"dynamics", "u850", "zonal_wind_at_850hPa", "second_diagnostics", "all", "none", "none", 2
"dynamics", "u850", "zonal_wind_at_850hPa", "second_diagnostics", "all", "none", "none", 2
"""
    assert str(diag_table) == expected_str


def test_DiagTable_as_dict(diag_table):
    expected_dict = {
        "name": "experiment",
        "base_time": datetime(2000, 1, 1, 0, 0),
        "file_configs": [
            {
                "name": "first_diagnostics",
                "frequency": 1,
                "frequency_units": "hours",
                "field_configs": [
                    {
                        "module_name": "dynamics",
                        "field_name": "u850",
                        "output_name": "zonal_wind_at_850hPa",
                        "reduction_method": "none",
                        "time_sampling": "all",
                        "regional_section": "none",
                        "packing": 2,
                    },
                    {
                        "module_name": "dynamics",
                        "field_name": "u850",
                        "output_name": "zonal_wind_at_850hPa",
                        "reduction_method": "none",
                        "time_sampling": "all",
                        "regional_section": "none",
                        "packing": 2,
                    },
                ],
                "file_format": 1,
                "time_axis_units": "hours",
                "time_axis_name": "time",
            },
            {
                "name": "second_diagnostics",
                "frequency": 1,
                "frequency_units": "hours",
                "field_configs": [
                    {
                        "module_name": "dynamics",
                        "field_name": "u850",
                        "output_name": "zonal_wind_at_850hPa",
                        "reduction_method": "none",
                        "time_sampling": "all",
                        "regional_section": "none",
                        "packing": 2,
                    },
                    {
                        "module_name": "dynamics",
                        "field_name": "u850",
                        "output_name": "zonal_wind_at_850hPa",
                        "reduction_method": "none",
                        "time_sampling": "all",
                        "regional_section": "none",
                        "packing": 2,
                    },
                ],
                "file_format": 1,
                "time_axis_units": "hours",
                "time_axis_name": "time",
            },
        ],
    }
    assert expected_dict == diag_table.asdict()


def test_from_str():
    input_str = """default_experiment
2016 08 01 00 0 0
#output files
"atmos_static",           -1,  "hours",    1, "hours", "time",
"atmos_dt_atmos",          2,  "hours",    1, "hours",  "time" #inline comment
"empty_file",              2,  "hours",    1, "hours",  "time"
#
 #output variables
#

###
# atmos_static
###
  "dynamics",      "zsurf",       "HGTsfc",       "atmos_static",      "all", .false.,  "none", 2
###
# atmos_dt_atmos
###
"dynamics",  "us",          "UGRDlowest",    "atmos_dt_atmos", "all", .false.,  "none", 2, # inline comment
"dynamics",  "u850",        "UGRD850",       "atmos_dt_atmos", "all", .true.,  "none", 2
"""

    diag_table = DiagTable.from_str(input_str)
    assert diag_table.name == "default_experiment"
    assert diag_table.base_time == datetime(2016, 8, 1)
    assert len(diag_table.file_configs) == 3
    assert diag_table.file_configs[0] == DiagFileConfig(
        "atmos_static", -1, "hours", [DiagFieldConfig("dynamics", "zsurf", "HGTsfc")],
    )
    assert diag_table.file_configs[1] == DiagFileConfig(
        "atmos_dt_atmos",
        2,
        "hours",
        [
            DiagFieldConfig("dynamics", "us", "UGRDlowest"),
            DiagFieldConfig("dynamics", "u850", "UGRD850", reduction_method="average"),
        ],
    )


@pytest.mark.parametrize(
    "diag_table_str",
    [
        """default_experiment
2016 08 01 00 0 0
"dynamics", "zsurf", "HGTsfc", "atmos_static", "all", .false.,  "none", 2
"atmos_static", -1,  "hours", 1, "hours", "time"
""",
        """default experiment
2016 08 01 00 0 0
"atmos_static", -1,  "hours", 1, "hours", "time"
"dynamics", "zsurf", "HGTsfc", "atmos_static", "all", .false.,  "none", 2
""",
    ],
    ids=["field_defined_before_corresponding_file", "spaces_in_name"],
)
def test_from_str_raises_config_error(diag_table_str):
    with pytest.raises(ConfigError):
        DiagTable.from_str(diag_table_str)


@pytest.mark.parametrize(
    "input_,expected_output",
    [
        ('"name"', "name"),
        (".true.", "average"),
        (".TRUE.", "average"),
        (".false.", "none"),
        (".FALSE.", "none"),
        ("3", 3),
    ],
)
def test__str_to_token(input_, expected_output):
    output = DiagTable._str_to_token(input_)
    assert output == expected_output


@pytest.mark.parametrize(
    "token,expected_output", [("name", '"name"'), (3, "3")],
)
def test__token_to_str(token, expected_output):
    output = DiagTable._token_to_str(token)
    assert output == expected_output



import sys
import logging
import contextlib
import resource
import functools
import subprocess
import inspect
import multiprocessing
import os
import tempfile
import warnings
import json
from ..config import write_run_directory, get_n_processes, dump, load
from .. import filesystem

STDOUT_FILENAME = "stdout.log"
STDERR_FILENAME = "stderr.log"
CONFIG_OUT_FILENAME = "fv3config.yml"
MPI_FLAGS = [
    "--allow-run-as-root",
    "--use-hwthread-cpus",
    "--mca",
    "btl_vader_single_copy_mechanism",
    "none",
]
RUNFILE_ENV_VAR = "FV3CONFIG_DEFAULT_RUNFILE"

logger = logging.getLogger("fv3run")


def call_via_subprocess(module):
    def decorator(func):
        signature = inspect.signature(func)

        def main(argv):
            args, kwargs = json.loads(argv[1])
            func(*args, **kwargs)

        @functools.wraps(func)
        def command(*args, **kwargs) -> str:
            # check that args and kwargs match func
            # raises TypeError if not
            signature.bind(*args, **kwargs)

            serialized = json.dumps([args, kwargs])
            return ["python", "-m", module, serialized]

        func.main = main
        func.command = command
        return func

    return decorator


@call_via_subprocess("fv3config.fv3run._native_main")
def run_native(
    config_dict_or_location, outdir, runfile=None, capture_output: bool = True
):
    """Run the FV3GFS model with the given configuration.

    Copies the resulting directory to a target location. Will use the Google cloud
    storage key at ``$GOOGLE_APPLICATION_CREDENTIALS`` by default. Requires the
    fv3gfs-python package.

    Args:
        config_dict_or_location (dict or str): a configuration dictionary, or a
            location (local or on Google cloud storage) of a yaml file containing
            a configuration dictionary
        outdir (str): location to copy the resulting run directory
        runfile (str, optional): Python model script to use in place of the default.
        capture_output (bool, optional): If true, then the stderr and stdout
            streams will be redirected to the files `outdir/stderr.log` and `outdir/stdout.log`
            respectively.
    """
    _set_stacksize_unlimited()
    with _temporary_directory(outdir) as localdir:
        config_out_filename = os.path.join(localdir, CONFIG_OUT_FILENAME)
        # we need to write the dict to the run directory for archival and also load
        # the dict, it ends up being convenient to do both at once
        config_dict = _get_config_dict_and_write(
            config_dict_or_location, config_out_filename
        )
        write_run_directory(config_dict, localdir)
        if runfile is not None:
            filesystem.get_file(
                runfile, os.path.join(localdir, os.path.basename(runfile))
            )
        with _output_stream_context(localdir, capture_output) as (stdout, stderr):
            n_processes = get_n_processes(config_dict)
            _run_experiment(
                localdir,
                n_processes,
                runfile=runfile,
                mpi_flags=_add_oversubscribe_if_necessary(MPI_FLAGS, n_processes),
                stdout=stdout,
                stderr=stderr,
            )


def _set_stacksize_unlimited():
    try:
        resource.setrlimit(
            resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY)
        )
    except ValueError:
        warnings.warn(
            "could not remove stacksize limit, may run out of memory as a result"
        )


def _add_oversubscribe_if_necessary(mpi_flags, n_processes):
    try:
        cpu_count = multiprocessing.cpu_count()
        if cpu_count < n_processes:
            mpi_flags += ["--oversubscribe"]
    except NotImplementedError:
        warnings.warn(
            "could not determine cpu count, assuming number of processors"
            "is at least as many as number of MPI tasks"
        )
    return mpi_flags


@contextlib.contextmanager
def _temporary_directory(outdir):
    fs = filesystem.get_fs(outdir)
    if not filesystem.is_local_path(outdir):
        with tempfile.TemporaryDirectory() as tempdir:
            try:
                yield tempdir
            finally:
                logger.info("Copying output to %s", outdir)
                fs.makedirs(outdir, exist_ok=True)
                filesystem.put_directory(tempdir, outdir)
    else:
        fs.makedirs(outdir, exist_ok=True)
        yield outdir


def _captured_output_context(localdir):
    out_filename = os.path.join(localdir, STDOUT_FILENAME)
    err_filename = os.path.join(localdir, STDERR_FILENAME)
    with open(out_filename, "wb") as out_file, open(err_filename, "wb") as err_file:
        try:
            yield out_file, err_file
        except subprocess.CalledProcessError as e:
            logger.critical(
                "Experiment failed. " "Check %s and %s for logs.",
                STDOUT_FILENAME,
                STDERR_FILENAME,
            )
            raise e


def _uncaptured_output_context(localdir):
    try:
        yield sys.stdout, sys.stderr
    except subprocess.CalledProcessError as e:
        logger.critical("Experiment failed")
        raise e


@contextlib.contextmanager
def _output_stream_context(localdir: str, capture_output: bool):
    logger.info("running experiment")
    if capture_output:
        yield from _captured_output_context(localdir)
    else:
        yield from _uncaptured_output_context(localdir)


def _get_python_command(runfile):
    python_args = ["python3", "-m", "mpi4py"]
    if runfile is not None:
        python_args.append(os.path.basename(runfile))
    elif RUNFILE_ENV_VAR in os.environ:
        python_args.append(os.environ[RUNFILE_ENV_VAR])
    else:
        python_args += ["-m", "fv3gfs.run"]
    return python_args


def _run_experiment(
    dirname, n_processes, runfile, mpi_flags=None, stdout=None, stderr=None
):
    if mpi_flags is None:
        mpi_flags = []

    python_command = _get_python_command(runfile)
    logger.info("Running experiment in %s", dirname)
    subprocess.check_call(
        ["mpirun", "-n", str(n_processes)] + mpi_flags + python_command,
        cwd=dirname,
        stdout=stderr,
        stderr=stdout,
    )


def _get_config_dict_and_write(config_dict_or_location, config_out_filename):
    if isinstance(config_dict_or_location, dict):
        config_dict = config_dict_or_location
        with filesystem.open(config_out_filename, "w") as f:
            dump(config_dict, f)
    else:
        config_dict = _copy_and_load_config_dict(
            config_dict_or_location, config_out_filename
        )
    return config_dict


def _copy_and_load_config_dict(config_location, local_target_location):
    filesystem.get_file(config_location, local_target_location)
    with open(local_target_location, "r") as infile:
        config_dict = load(infile)
    return config_dict


if __name__ == "__main__":
    # In theory this warning should never be triggered.
    # There's probably a bug in run_native.command if it is.
    # Remove this main block after some time if it never gets triggered.
    warnings.warn(
        "calling fv3config.fv3run._native is deprecated, call fv3config.fv3run._native_main instead",
        DeprecationWarning,
    )
    run_native.main(sys.argv)



import pytest
from datetime import datetime, timedelta

from fv3config.config import nudging
import fv3config


@pytest.fixture
def test_config(c12_config):
    return c12_config


@pytest.mark.parametrize(
    "start_time, interval, expected",
    [
        (datetime(2016, 1, 1), timedelta(hours=6), datetime(2016, 1, 1, 0)),
        (datetime(2016, 1, 1, 1), timedelta(hours=6), datetime(2016, 1, 1, 0)),
        (datetime(2016, 1, 1, 1), timedelta(hours=2), datetime(2016, 1, 1, 0)),
        (datetime(2016, 1, 1, 3), timedelta(hours=2), datetime(2016, 1, 1, 2)),
        (datetime(2016, 1, 1, 7), timedelta(hours=6), datetime(2016, 1, 1, 6)),
        (datetime(2016, 1, 1, 12), timedelta(hours=6), datetime(2016, 1, 1, 12)),
        (datetime(2016, 1, 2, 18, 1), timedelta(hours=6), datetime(2016, 1, 2, 18)),
    ],
)
def test__get_first_nudge_file_time(start_time, interval, expected):
    assert nudging._most_recent_nudge_time(start_time, interval) == expected


@pytest.mark.parametrize(
    "duration, current_date, interval, expected_length, "
    "expected_first_datetime, expected_last_datetime",
    [
        (
            timedelta(days=1),
            [2016, 1, 1, 0, 0, 0],
            timedelta(hours=6),
            4 + 1,
            datetime(2016, 1, 1),
            datetime(2016, 1, 2),
        ),
        (
            timedelta(days=1),
            [2016, 1, 1, 0, 0, 0],
            timedelta(hours=2),
            12 + 1,
            datetime(2016, 1, 1),
            datetime(2016, 1, 2),
        ),
        (
            timedelta(days=1, hours=5),
            [2016, 1, 1, 0, 0, 0],
            timedelta(hours=6),
            4 + 1 + 1,
            datetime(2016, 1, 1),
            datetime(2016, 1, 2, 6),
        ),
        (
            timedelta(days=1, hours=7),
            [2016, 1, 1, 0, 0, 0],
            timedelta(hours=6),
            4 + 2 + 1,
            datetime(2016, 1, 1),
            datetime(2016, 1, 2, 12),
        ),
        (
            timedelta(days=1),
            [2016, 1, 2, 1, 0, 0],
            timedelta(hours=6),
            4 + 2,
            datetime(2016, 1, 2),
            datetime(2016, 1, 3, 6),
        ),
    ],
)
def test__get_nudge_time_list(
    duration,
    current_date,
    interval,
    expected_length,
    expected_first_datetime,
    expected_last_datetime,
):
    nudge_file_list = nudging._get_nudge_time_list(duration, current_date, interval)
    assert len(nudge_file_list) == expected_length
    assert nudge_file_list[0] == expected_first_datetime
    assert nudge_file_list[-1] == expected_last_datetime


def test_get_nudging_assets():
    nudge_url = "/path/to/files"
    pattern = "%Y%m%d_%H.nc"
    assets = fv3config.get_nudging_assets(
        timedelta(hours=6),
        [2016, 1, 1, 0, 0, 0],
        nudge_url,
        nudge_filename_pattern=pattern,
    )
    expected = [
        fv3config.get_asset_dict(nudge_url, "20160101_00.nc", target_location="INPUT"),
        fv3config.get_asset_dict(nudge_url, "20160101_06.nc", target_location="INPUT"),
    ]
    assert assets == expected


def test_get_nudging_assets_raises_config_error_if_linking_remote_files():
    with pytest.raises(fv3config.ConfigError):
        fv3config.get_nudging_assets(
            timedelta(hours=6),
            [2016, 1, 1, 0, 0, 0],
            "gs://bucket/path",
            copy_method="link",
        )


@pytest.mark.parametrize(
    "nudging_assets, other_items, pattern",
    [
        ([], [], "%Y%m%d_%H.nc"),
        ([fv3config.get_asset_dict("/path", "20160101_00.nc")], [], "%Y%m%d_%H.nc"),
        ([], ["non asset item"], "%Y%m%d_%H.nc"),
        ([], [{"other": "non asset item"}], "%Y%m%d_%H.nc"),
        (
            [fv3config.get_asset_dict("/path", "20160101_00.nc")],
            ["non asset item"],
            "%Y%m%d_%H.nc",
        ),
        (
            [fv3config.get_asset_dict("/path", "20160101_00.nc")],
            [fv3config.get_asset_dict("/path", "non nudging asset"), "non asset"],
            "%Y%m%d_%H.nc",
        ),
    ],
)
def test__non_nudging_assets(nudging_assets, other_items, pattern):
    input_assets = nudging_assets + other_items
    cleared_assets = nudging._non_nudging_assets(input_assets, pattern)
    for item in nudging_assets:
        assert item not in cleared_assets
    for item in other_items:
        assert item in cleared_assets


@pytest.mark.parametrize(
    "item, pattern, expected",
    [
        (fv3config.get_asset_dict("/path", "20160101_00.nc"), "%Y%m%d_%H.nc", True),
        (fv3config.get_asset_dict("/path", "not_nudging_file"), "%Y%m%d_%H.nc", False),
    ],
)
def test__is_nudging_asset(item, pattern, expected):
    assert nudging._is_nudging_asset(item, pattern) == expected


def test_update_config_for_nudging(test_config):
    url = "/path/to/nudging/files"
    pattern = "%Y%m%d_%H.nc"
    old_nudging_file = "20151231_18.nc"
    new_nudging_file = "20160101_06.nc"
    old_asset = fv3config.get_asset_dict(url, old_nudging_file, target_location="INPUT")
    new_asset = fv3config.get_asset_dict(url, new_nudging_file, target_location="INPUT")
    test_config["gfs_analysis_data"] = {"url": url, "filename_pattern": pattern}
    test_config["namelist"]["coupler_nml"]["current_date"] = [2016, 1, 1, 0, 0, 0]
    test_config["namelist"]["coupler_nml"]["hours"] = 12
    test_config["namelist"]["fv_nwp_nudge_nml"] = {
        "file_names": [f"INPUT/{old_nudging_file}"]
    }
    test_config["patch_files"] = [old_asset]

    fv3config.update_config_for_nudging(test_config)

    updated_file_names = test_config["namelist"]["fv_nwp_nudge_nml"]["file_names"]
    assert f"INPUT/{old_nudging_file}" not in updated_file_names
    assert f"INPUT/{new_nudging_file}" in updated_file_names
    assert old_asset not in test_config["patch_files"]
    assert new_asset in test_config["patch_files"]



import subprocess
import os
from .. import filesystem
from ..config._serialization import load
from ._native import CONFIG_OUT_FILENAME, run_native

DOCKER_OUTDIR = "/outdir"
DOCKER_CONFIG_LOCATION = os.path.join("/", CONFIG_OUT_FILENAME)
DOCKER_RUNFILE = "/runfile.py"
DOCKER_COMMAND = ["docker", "run"]
DOCKER_KEYFILE = "/gcs_key.json"
FV3RUN_MODULE = "fv3config.fv3run"


def run_docker(
    config_dict_or_location,
    outdir,
    docker_image,
    runfile=None,
    keyfile=None,
    capture_output=True,
):
    """Run the FV3GFS model in a docker container with the given configuration.

    Copies the resulting directory to a target location. Will use the Google cloud
    storage key at ``$GOOGLE_APPLICATION_CREDENTIALS`` by default. Requires the
    fv3gfs-python package and fv3config to be installed in the docker image.

    Args:
        config_dict_or_location (dict or str): a configuration dictionary, or a
            location (local or on Google cloud storage) of a yaml file containing
            a configuration dictionary
        outdir (str): location to copy the resulting run directory
        runfile (str, optional): Python model script to use in place of the default.
        docker_image (str, optional): If given, run this command inside a container
            using this docker image. Image must have this package and fv3gfs-python
            installed.
        keyfile (str, optional): location of a Google cloud storage key to use
            inside the docker container
        capture_output (bool, optional): If true, then the stderr and stdout
            streams will be redirected to the files `outdir/stderr.log` and `outdir/stdout.log`
            respectively.
    """
    if keyfile is None:
        keyfile = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS", None)

    if isinstance(config_dict_or_location, str):
        config_dict = _load_yaml(config_dict_or_location)
    else:
        config_dict = config_dict_or_location

    filesystem.get_fs(outdir).makedirs(outdir, exist_ok=True)
    bind_mount_args = []
    docker_args = []
    _get_credentials_args(keyfile, docker_args, bind_mount_args)
    _get_local_data_bind_mounts(config_dict, bind_mount_args)

    _get_docker_args(docker_args, bind_mount_args, outdir)
    runfile_in_docker = _get_runfile_args(runfile, bind_mount_args)
    if filesystem.is_local_path(outdir):
        docker_outdir = DOCKER_OUTDIR
    else:
        docker_outdir = outdir
    python_command = run_native.command(
        config_dict,
        docker_outdir,
        runfile=runfile_in_docker,
        capture_output=capture_output,
    )

    subprocess.check_call(
        DOCKER_COMMAND + bind_mount_args + docker_args + [docker_image] + python_command
    )


def _load_yaml(url):
    with filesystem.open(url) as f:
        return load(f)


def _get_runfile_args(runfile, bind_mount_args) -> str:
    if runfile is not None:
        if filesystem.is_local_path(runfile):
            bind_mount_args += ["-v", f"{os.path.abspath(runfile)}:{DOCKER_RUNFILE}"]
            return DOCKER_RUNFILE
        else:
            return runfile


def _get_paths(config_dict):
    """Return a list of all paths referenced by the config dict."""
    return_list = [
        config_dict["diag_table"],
        config_dict["data_table"],
        config_dict["forcing"],
        config_dict["initial_conditions"],
    ]
    patch_files = config_dict.get("patch_files", [])
    if isinstance(patch_files, list):
        return_list.extend(patch_files)
    else:
        return_list.append(patch_files)
    return return_list


def _is_local_path(maybe_path_or_object):
    return (
        isinstance(maybe_path_or_object, str)
        and os.path.isabs(maybe_path_or_object)
        and filesystem.is_local_path(maybe_path_or_object)
    )


def _get_local_data_paths(config_dict):
    """Return a list of all local paths referenced by the config dict."""
    local_paths = []
    for potential_path in _get_paths(config_dict):
        if _is_local_path(potential_path):
            local_paths.append(potential_path)
        elif isinstance(potential_path, list):
            for asset in potential_path:
                if filesystem.is_local_path(asset["source_location"]):
                    local_paths.append(
                        os.path.join(asset["source_location"], asset["source_name"])
                    )
        elif isinstance(potential_path, dict):
            asset = potential_path
            if filesystem.is_local_path(asset["source_location"]):
                local_paths.append(
                    os.path.join(asset["source_location"], asset["source_name"])
                )
    return local_paths


def _get_local_data_bind_mounts(config_dict, bind_mount_args):
    for local_path in _get_local_data_paths(config_dict):
        bind_mount_args += ["-v", f"{local_path}:{local_path}"]


def _get_docker_args(docker_args, bind_mount_args, outdir):
    if filesystem.is_local_path(outdir):
        bind_mount_args += ["-v", f"{os.path.abspath(outdir)}:{DOCKER_OUTDIR}"]
    docker_args += ["--rm", "--user", f"{os.getuid()}:{os.getgid()}"]


def _get_credentials_args(keyfile, docker_args, bind_mount_args):
    if keyfile is not None:
        bind_mount_args += ["-v", f"{os.path.abspath(keyfile)}:{DOCKER_KEYFILE}"]
        docker_args += ["-e", f"GOOGLE_APPLICATION_CREDENTIALS={DOCKER_KEYFILE}"]



import sys
import argparse
from ._docker import run_docker
from ._kubernetes import run_kubernetes
from ._native import run_native, RUNFILE_ENV_VAR
import yaml
import logging

MODULE_NAME = "fv3config.run"
STDOUT_FILENAME = "stdout.log"
STDERR_FILENAME = "stderr.log"
DOCKER_FLAGS = "-it"


def _parse_args():
    parser = argparse.ArgumentParser(
        description="""Run the FV3GFS model.

Will use google cloud storage key at $GOOGLE_APPLICATION_CREDENTIALS by default.
"""
    )
    parser.add_argument(
        "config", type=str, action="store", help="location of fv3config yaml file"
    )
    parser.add_argument(
        "outdir",
        type=str,
        action="store",
        help="location to copy final run directory, used as run directory if local",
    )
    parser.add_argument(
        "--runfile",
        type=str,
        action="store",
        help="Location of python script to execute with mpirun. If not specified, a "
        f"default is used, which can be overriden by setting the {RUNFILE_ENV_VAR}.",
    )
    parser.add_argument(
        "--dockerimage",
        type=str,
        action="store",
        help="if passed, execute inside a docker image with the given name",
    )
    parser.add_argument(
        "--keyfile",
        type=str,
        action="store",
        help="google cloud storage key to use for cloud copy commands",
    )
    parser.add_argument(
        "--kubernetes",
        action="store_true",
        default=False,
        help=(
            "if given, ignore --keyfile and output a yaml kubernetes config to stdout "
            "instead of submitting a run"
        ),
    )
    parser.add_argument(
        "--capture-output",
        action="store_true",
        default=False,
        help="If given, save the outputs of the fv3gfs call in a outdir/stderr.log and "
        "outdir/stdout.log. Not recommended for use with docker or kubernetes. "
        "It is recommended to use default linux pipes or docker's and kuberentes' logging "
        "functionality.",
    )
    return parser.parse_args()


def main():
    """Run the FV3GFS model based on a configuration dictionary.
    Copies the resulting run directory to a target location.
    """
    args = _parse_args()
    logging.basicConfig(level=logging.INFO)

    if args.dockerimage is not None:
        if args.kubernetes:
            job = run_kubernetes(
                args.config,
                args.outdir,
                args.dockerimage,
                runfile=args.runfile,
                submit=False,
            )
            yaml.dump(job.to_dict(), stream=sys.stdout)
        else:
            run_docker(
                args.config,
                args.outdir,
                args.dockerimage,
                runfile=args.runfile,
                keyfile=args.keyfile,
                capture_output=args.capture_output,
            )
    else:
        run_native(
            args.config,
            args.outdir,
            runfile=args.runfile,
            capture_output=args.capture_output,
        )


if __name__ == "__main__":
    sys.exit(main())



import fv3config
import pytest
from datetime import timedelta


@pytest.fixture
def config(c12_config):
    return c12_config


@pytest.mark.parametrize(
    "coupler_nml, duration",
    [
        ({"seconds": 10}, timedelta(seconds=10)),
        ({"minutes": 2}, timedelta(minutes=2)),
        ({"hours": 100}, timedelta(hours=100)),
        ({"days": 9}, timedelta(days=9)),
        ({"months": 6}, None),
        ({"seconds": 5, "minutes": 1, "hours": 1}, timedelta(seconds=65, hours=1)),
    ],
)
def test_get_run_duration(coupler_nml, duration, config):
    config["namelist"]["coupler_nml"] = coupler_nml
    if coupler_nml.get("months", 0) != 0:
        with pytest.raises(ValueError):
            fv3config.get_run_duration(config)
    else:
        assert fv3config.get_run_duration(config) == duration



# -*- coding: utf-8 -*-

"""Top-level package for fv3config."""

from .config import (
    config_to_namelist,
    config_from_namelist,
    get_default_config,
    write_run_directory,
    enable_restart,
    get_run_duration,
    set_run_duration,
    get_timestep,
    get_nudging_assets,
    update_config_for_nudging,
    DiagTable,
    DiagFieldConfig,
    DiagFileConfig,
    Packing,
    FileFormat,
    dump,
    load,
)
from ._exceptions import InvalidFileError, ConfigError
from ._datastore import ensure_data_is_downloaded
from .fv3run import run_docker, run_native, run_kubernetes
from ._asset_list import get_asset_dict, get_bytes_asset_dict, asset_list_from_path
from .caching import CACHE_REMOTE_FILES, do_remote_caching, set_cache_dir, get_cache_dir


__author__ = """Vulcan Technologies LLC"""
__email__ = "jeremym@vulcan.com"
__version__ = "0.7.1"

# necessary for auto-doc generation of API for public methods
__all__ = dir()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

with open("requirements.txt") as requirements_file:
    requirements = requirements_file.read().splitlines()

setup_requirements = []

test_requirements = ["pytest"]

setup(
    author="Vulcan Technologies LLC",
    author_email="jeremym@vulcan.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="FV3Config is used to configure and manipulate run directories for FV3GFS.",
    entry_points={
        "console_scripts": [
            "fv3run=fv3config.fv3run.__main__:main",
            "write_run_directory=fv3config.cli:write_run_directory",
            "enable_restart=fv3config.cli:enable_restart",
            "update_config_for_nudging=fv3config.cli:update_config_for_nudging",
        ]
    },
    install_requires=requirements,
    extras_require={
        "bucket-access": "gcsfs",
        "fv3run": "fv3gfs-python",
        "run_kubernetes": "kubernetes",
    },
    license="BSD license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="fv3config",
    name="fv3config",
    packages=find_packages(include=["fv3config", "fv3config.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/VulcanClimateModeling/fv3config",
    version="0.7.1",
    zip_safe=False,
)



import os
import fsspec
import backoff
from ._exceptions import DelayedImportError
from . import caching
from concurrent.futures import ThreadPoolExecutor, Executor


try:
    import gcsfs

    FSSPEC_ERRORS = (RuntimeError, gcsfs.utils.HttpError, gcsfs.utils.ChecksumError)
except ImportError as err:
    gcsfs = DelayedImportError(err)
    FSSPEC_ERRORS = RuntimeError
try:
    import google.auth
except ImportError as err:
    google = DelayedImportError(err)


fsspec_backoff = backoff.on_exception(backoff.expo, FSSPEC_ERRORS, max_time=60)


def get_fs(path: str) -> fsspec.AbstractFileSystem:
    """Return the fsspec filesystem required to handle a given path."""
    return _get_fs(path)


def _get_fs(path: str) -> fsspec.AbstractFileSystem:
    """Private function implementing public get_fs function, used so that if we
    mock this implementation, it is still used in modules which import using
    `from filesystem import get_fs`.
    """
    if path.startswith("gs://"):
        return fsspec.filesystem("gs", requester_pays=True)
    else:
        return fsspec.filesystem("file")


def isabs(path: str) -> bool:
    """Return whether the path is a local or remote absolute path"""
    if len(_get_protocol_prefix(path)) > 0:
        return True
    else:  # local file
        return os.path.isabs(path)


def is_existing_absolute_path(path: str) -> bool:
    """Return whether the path is an existing absolute path"""
    return isabs(path) and get_fs(path).exists(path)


def _get_protocol_prefix(location):
    """If a string starts with "<protocol>://"", return that part of the string.
    Otherwise, return an empty string.
    """
    separator = "://"
    if separator in location:
        return location[: location.index(separator) + len(separator)]
    else:
        return ""


def _get_path(location):
    """If a string starts with "<protocol>://"", return the rest of the string.
    Otherwise, return the entire string.
    """
    separator = "://"
    if separator in location:
        return location[location.index(separator) + len(separator) :]
    else:
        return location


def is_local_path(location):
    """returns True if the location is local, False otherwise"""
    return _get_protocol_prefix(location) == ""


def put_directory(
    local_source_dir: str,
    dest_dir: str,
    fs: fsspec.AbstractFileSystem = None,
    executor: Executor = None,
):
    """Copy the contents of a local directory to a local or remote directory.
    """
    if fs is None:
        fs = get_fs(dest_dir)
    if executor is None:
        executor = ThreadPoolExecutor()
        manage_threads = True
    else:
        manage_threads = False
    for token in os.listdir(local_source_dir):
        source = os.path.join(os.path.abspath(local_source_dir), token)
        dest = os.path.join(dest_dir, token)
        if os.path.isdir(source):
            fsspec_backoff(fs.makedirs)(dest, exist_ok=True)  # must be blocking call
            put_directory(source, dest, fs=fs, executor=executor)
        else:
            executor.submit(fsspec_backoff(fs.put), source, dest)
    if manage_threads:
        executor.shutdown(wait=True)


def get_file(source_filename: str, dest_filename: str, cache: bool = None):
    """Copy a file from a local or remote location to a local location.

    Optionally cache remote files in the local fv3config cache.
    
    Args:
        source_filename: the local or remote location to copy
        dest_filename: the local target location
        cache (optional): if True and source is remote, copy the file from the
            fv3config cache if it has been previously downloaded, and cache the file
            if not. Does nothing if source_filename is local.
            Default ``fv3config.caching.CACHE_REMOTE_FILES``, set by
            ``fv3config.enable_remote_caching(True/False)``.
    """
    if cache is None:
        cache = caching.CACHE_REMOTE_FILES
    if not cache or is_local_path(source_filename):
        _get_file_uncached(source_filename, dest_filename)
    else:
        _get_file_cached(source_filename, dest_filename)


@fsspec_backoff
def _get_file_uncached(source_filename, dest_filename):
    fs = get_fs(source_filename)
    fs.get(source_filename, dest_filename)


def _get_file_cached(source_filename, dest_filename):
    if is_local_path(source_filename):
        raise ValueError(f"will not cache a local path, was given {source_filename}")
    else:
        cache_location = _get_cache_filename(source_filename)
        if not os.path.isfile(cache_location):
            os.makedirs(os.path.dirname(cache_location), exist_ok=True)
            _get_file_uncached(source_filename, cache_location)
        _get_file_uncached(cache_location, dest_filename)


@fsspec_backoff
def put_file(source_filename, dest_filename):
    """Copy a file from a local location to a local or remote location.
    
    Args:
        source_filename (str): the local location to copy
        dest_filename (str): the local or remote target location
    """
    fs = get_fs(dest_filename)
    fs.put(source_filename, dest_filename)


def _get_cache_filename(source_filename):
    prefix = _get_protocol_prefix(source_filename).strip("://")
    path = _get_path(source_filename)
    if len(path) == 0:
        raise ValueError(f"no file path given in source filename {source_filename}")
    cache_dir = caching.get_internal_cache_dir()
    return os.path.join(cache_dir, prefix, path)


open = fsspec.open
